package javamath.hasse.lattice;

/* OrderedSet.java	96/20/6

I noted that the lisp code found a linear extension by taking the minimal
elements then the minimal of what remains, etc. I had assumed that this
would order it with increasing ranks (of my rank function).
It doesn't so I calculate the ranks then do a quicksort to rearrange
elems. Then I do a setElemOrder a second time.

The language specification may not guarentee that v.elements() has
the same order as the Vector v. 
I checked: it does guarentee that it has the right order. 

I noted that the lisp code for finding filter from and edge set E
did not truly implement the fast algorithm of "Free Lattices" since
the uppers of each element were not topologically sorted. The algorithm 
is still correct, just not guaranteed to be as fast. Since this only matters
if E properly contains E_cov this is not important for the lisp
program since that rarely happens. 

A fast way to sort the upper covers lists is to first find a linear
extension of P then covert the upper covers to lower covers and then
back again using the lin ext of P.


*/



import java.util.*;
import java.io.*;
import javamath.hasse.util.MyInt;
import javamath.hasse.util.LinkedList;
import javamath.hasse.util.EmptyListException;
import javamath.hasse.util.QuickSort;


public class OrderedSet {

  Vector elems;			// topologically sorted
  Hashtable elems_ht;		// labels to elems
  Hashtable elemOrder;		// labels to Integers
  //Hashtable filter;		// labels to filters, a Vector 
  boolean[][] leqTable;


	// AS: begin name stuff
	String name = null;

	public void setName(String n)
	{
		name = n;
	}


	public String getName()
	{
		return name;
	}
	// AS: end name stuff

  // Constructors
	// AS: added <name>
  public OrderedSet(String name, Vector labels, Vector ucs) 
		throws NonOrderedSetException 
	{
		// AS: added <name>
		if (name !=null)
		{
			setName(name);
		}

    if (labels.size() == 0) 
		{
       throw new NonOrderedSetException(NonOrderedSetException.emptyError);
    }
    linExt(labels, ucs);
    setElemOrder();
    setFilters();
    setLeq();
    setIdeals();
    setUpperCoversFromFilters();
    setLowerCovers();
    setRanks();
    new QuickSortElems(elems);
    setElemOrder();
    setLeq();
    setIncomparables();
  }

	// AS: added <name>
  public OrderedSet(InputLattice in) throws NonOrderedSetException {
    this(in.name, in.labels, in.upperCoversVector);
  }

  // public methods

  public int card() {
    return elems.size();
  }

  public Vector univ() {
    return elems;
  }

  public POElem one() {
    return (POElem)elems.elementAt(card() - 1);
  }

  public POElem zero() {
    return (POElem)elems.elementAt(0);
  }

  public POElem getElement(String label) {
    return (POElem)elems_ht.get(label);
  }

  public int elemOrder(POElem elem) {
    return ((Integer)elemOrder.get(elem.label())).intValue();
  }

  public boolean leq(POElem x, POElem y) {
    return leqTable[elemOrder(x)][elemOrder(y)];
    //boolean ans = leqTable[elemOrder(x)][elemOrder(y)];
    //System.out.println("testing leq(" + elemOrder(x) + ", " + elemOrder(y) + 
		//") = " + ans);
    //return ans;
  }

  public boolean lt(POElem x, POElem y) {
    return elemOrder(x) != elemOrder(y) && leqTable[elemOrder(x)][elemOrder(y)];
  }

  public boolean geq(POElem x, POElem y) {
    return leqTable[elemOrder(y)][elemOrder(x)];
  }

  public boolean gt(POElem x, POElem y) {
    return elemOrder(x) != elemOrder(y) && leqTable[elemOrder(y)][elemOrder(x)];
  }

  //  helpers

  void linExt(Vector labels, Vector ucs) throws NonOrderedSetException {
    int n = labels.size();
    //Stack S = new Stack();
    Stack Z = new Stack();
    Stack ZNew = new Stack();
    Vector ans = new Vector(n);
    Hashtable uc = new Hashtable(n);
    Hashtable in = new Hashtable(n);
    for (int i = 0; i < n; i++) {
      in.put(labels.elementAt(i),new MyInt(0));
    }
    for (int i = 0; i < n; i++) {
      String a = (String)labels.elementAt(i);
      Vector upperCovers_a = (Vector)ucs.elementAt(i);
      int k = upperCovers_a.size();
      uc.put(a, upperCovers_a);
      for (int j=0; j < k; j++) {
	String b = (String)upperCovers_a.elementAt(j);
	if (! a.equals(b)) {
          ((MyInt)in.get(b)).increment();
	}
      }
    }
    Enumeration list = labels.elements();
    while (list.hasMoreElements()) {
      String a = (String)list.nextElement();
      if (0 == ((MyInt)in.get(a)).value()) {
	Z.push(a);
      }
    }
    while (true) {
      if (Z.empty()) {
	if (ZNew.empty()) {
	  break;
	} else {
	  Z = ZNew;
	  ZNew = new Stack();
	}
      }
      String a = (String)Z.pop();
      ans.addElement(a);
      Enumeration uc_a = ((Vector)uc.get(a)).elements();
      while (uc_a.hasMoreElements()) {
	String b = (String)uc_a.nextElement();
	if (! a.equals(b)) ((MyInt)in.get(b)).decrement();
	if (0 == ((MyInt)in.get(b)).value()) ZNew.push(b);
      }
    }
    if (ans.size() != n) {
	//System.out.println("This is not an ordered set.");
	//return;
        throw new NonOrderedSetException();
    }
    else {
      elems = new Vector(n);
      elems_ht = new Hashtable();   
      for (int i = 0; i < n; i++) {
	String label = (String)ans.elementAt(i);
        POElem elem = new POElem(label);
	elems.addElement(elem);
	elems_ht.put(label,elem);
      }
      for (int i = 0; i < n; i++) {
	POElem elem = (POElem)elems.elementAt(i);
	Vector up_covs = (Vector)uc.get(elem.label);
        int k = up_covs.size();
	Vector ucs2 = new Vector(k);
        for (int j = 0; j < k; j++) {
	  ucs2.addElement(elems_ht.get(up_covs.elementAt(j)));
	}
	elem.upperCovers = ucs2;
      }
      this.elems = elems;
      this.elems_ht = elems_ht;
    }
  }

  void setElemOrder() {
    elemOrder = new Hashtable(this.card());
    Enumeration list = elems.elements();
    int k = 0;
    while (list.hasMoreElements()) {
      elemOrder.put(((POElem)list.nextElement()).label(), new Integer(k++));
    }
  }

/* 
   g is an array of size k (the number of chains in the chain dec).
   Each element of the array is a hashtable. 
   The keys of the hashtables are the labels of the elements of the 
   ordered set.
   The values are a LinkedList with those elements in chain i above x.

   The algorithm for this is Listing 11.5 of "Free Lattices."
*/

  void setFilters() {
    Enumeration uc_x;
    POElem x,y,x_h,y_h;
    int n = this.card();
    Vector elemsRev = new Vector(n);
    ChainDecomposition chainDec = new ChainDecomposition(this);
    int k = chainDec.numChains();
    Hashtable[] g = new Hashtable[k];
    Enumeration list; 
    LinkedList llist;
    int index;
    Enumeration list_xh;
    for(int h = 0; h < k; h++) {
      g[h] = new Hashtable();
    }
    // reverse the order of the elements:
    for (int i = n-1; i >= 0; i--) {
      elemsRev.addElement(elems.elementAt(i));
    }
    list = elemsRev.elements();
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      for(int h = 0; h < k; h++) {
	g[h].put(x.label(), new LinkedList());
      }
    }
    list = elemsRev.elements();
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      uc_x = x.upperCovers().elements();
      while (uc_x.hasMoreElements()) {
	y = (POElem)uc_x.nextElement();
        if (x == y) continue;
	for(int h=0; h < k; h++) {
	  if (g[h].containsKey(y.label()) && 
		(! ((LinkedList)g[h].get(y.label())).isEmpty())) {
	    if ((! g[h].containsKey(x.label())) || 
		((LinkedList)g[h].get(x.label())).isEmpty()) {
	      g[h].put(x.label(), g[h].get(y.label()));
            } else {
              try { 
                x_h = (POElem)((LinkedList)g[h].get(x.label())).first();
                y_h = (POElem)((LinkedList)g[h].get(y.label())).first();
              } catch (EmptyListException e) {
                throw new InternalError(e.toString());
              }
              if (elemOrder(y_h) < elemOrder(x_h)) {
	        g[h].put(x.label(), g[h].get(y.label()));
              }
	    }
	  }
	}
      }
      index = chainDec.chainNum(x);
      llist = new LinkedList(x, (LinkedList)g[index].get(x.label()));
      g[index].put(x.label(), llist);
    }
    list = elemsRev.elements();
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      Vector filter_x = new Vector();
      // Make sure the chain with x is the first one processed
      // so x is the first element of x.filter.
      index = chainDec.chainNum(x);
      list_xh = 
        ((LinkedList)g[index].get(x.label())).elements();
      while (list_xh.hasMoreElements()) {
        filter_x.addElement(list_xh.nextElement());
      }
      for(int h = 0; h < k; h++) {
        if (index != h) {
	  list_xh = ((LinkedList)g[h].get(x.label())).elements();
	  while (list_xh.hasMoreElements()) {
	    filter_x.addElement(list_xh.nextElement());
          }
 	}
      }
      x.filter = filter_x;
    }
  }

  void setLeq () {
    int n = card();
    boolean[][] table = new boolean[n][n];
    Enumeration list = univ().elements();
    Enumeration list2; 
    POElem x;
    int k,j;
    Vector filter;
    for (k=0; k < n; k++) {
      for (j=0; j < n; j++) {
	table[k][j] = false;
      }
    }
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      k = elemOrder(x);
      filter = x.filter();
      list2 = filter.elements();
      while (list2.hasMoreElements()) {
	j = elemOrder((POElem)list2.nextElement());
	table[k][j] = true;
      }
    }
    leqTable = table;
  }

  void setIdeals() {
    int n = card();
    Vector[] a = new Vector[n];
    Enumeration list = elems.elements();
    Enumeration list2;
    int k;
    POElem x;
    for (int i = 0; i < n; i++) {
      a[i] = new Vector();
    }
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      list2 = x.filter().elements();
      while (list2.hasMoreElements()) {
        k =  elemOrder((POElem)list2.nextElement());
        a[k].addElement(x);
      }
    }
    for(int i = 0; i < n; i++) {
      x = (POElem)elems.elementAt(i);
      x.ideal = a[i];
    }
  }

  void setLowerCovers() {
    int n = card();
    Vector[] a = new Vector[n];
    Enumeration list = elems.elements();
    Enumeration list2;
    int k;
    POElem x;
    for (int i = 0; i < n; i++) {
      a[i] = new Vector();
    }
    while (list.hasMoreElements()) {
      x = (POElem)list.nextElement();
      list2 = x.upperCovers().elements();
      while (list2.hasMoreElements()) {
        k =  elemOrder((POElem)list2.nextElement());
        a[k].addElement(x);
      }
    }
    for(int i = 0; i < n; i++) {
      x = (POElem)elems.elementAt(i);
      x.lowerCovers = a[i];
    }
  }

  //This follows the Listing 11.8 on page 215 of Free Lattices

  void setUpperCoversFromFilters () {
    int n = card();
    Vector[] ucs = new Vector[n];
    Vector uc;
    Enumeration T;
    int k, m;
    POElem a, x;
    for (int i = 0; i < n; i++) { 
      ucs[i] = new Vector();
    }
    for (int i = 0; i < n; i++) {
      a = (POElem)elems.elementAt(i);
      k =  elemOrder(a);			// elemOrder of a
      for (int j = i + 1; j < n; j++) {
        x =  (POElem)elems.elementAt(j);
        if (leq(a,x)) {
          uc = ucs[k];
          m = uc.size();
          int p = 0;
          while (p < m && ! leq((POElem)uc.elementAt(p), x)) {
            p++;
          }
          if (p == m) uc.addElement(x);
        }
      }
    }
    for(int i = 0; i < n; i++) {
      a = (POElem)elems.elementAt(i);
      a.upperCovers = ucs[i];
    }
  }

  void setRanks () {
    int n = card();
    int[] heights = new int[n];
    int[] depths = new int[card()];
    int height, depth, height2, depth2;
    int max_depth = 0;
    Vector lcs, ucs;
    Enumeration list;
    for(int i=0; i < n; i++) {
      height = 0;
      lcs = ((POElem)elems.elementAt(i)).lowerCovers();
      list = lcs.elements();
      while (list.hasMoreElements()) {
	height2 = 1 + heights[elemOrder((POElem)list.nextElement())];
	if (height2 > height) height = height2;
      }
      heights[i] = height;
    }
    for(int i=n-1; i >= 0; i--) {
      depth = 0;
      ucs = ((POElem)elems.elementAt(i)).upperCovers();
      list = ucs.elements();
      while (list.hasMoreElements()) {
	depth2 = 1 + depths[elemOrder((POElem)list.nextElement())];
	if (depth2 > depth) depth = depth2;
      }
      depths[i] = depth;
      if (depth > max_depth) max_depth = depth;
    }
    for(int i=0; i < n; i++) {
      ((POElem)elems.elementAt(i)).rank = max_depth + heights[i] - depths[i];
    }
  }

  // x.highIncomparables will be a list of those elements incomparble with x
  // which come after it in the linear order.

  void setIncomparables() {
    int n = elems.size();
    for (int i = 0 ; i < n - 1; i++) {
      POElem x = (POElem)elems.elementAt(i);
      LinkedList list = new LinkedList();
      for (int j = i + 1; j < n; j++) {
        if (! leq(x, (POElem)elems.elementAt(j))) list.push(elems.elementAt(j));
      }
      x.highIncomparables = list;
    }
  }
	
  public static void main(String args[]) 
		throws FileNotFoundException,IOException {
    if (args.length == 0) {
      Vector elems = new Vector(5);
      Vector ucs = new Vector(5);
      elems.addElement("0");
      elems.addElement("a");
      elems.addElement("b");
      elems.addElement("c");
      elems.addElement("1");
      Vector cov_0 = new Vector(2);
      Vector cov_a = new Vector(1);
      Vector cov_b = new Vector(1);
      Vector cov_c = new Vector(1);
      Vector cov_1 = new Vector(0);
      cov_0.addElement("b");
      cov_0.addElement("c");
      cov_a.addElement("1");
      cov_b.addElement("1");
      cov_c.addElement("a");
      ucs.addElement(cov_0);
      ucs.addElement(cov_a);
      ucs.addElement(cov_b);
      ucs.addElement(cov_c);
      ucs.addElement(cov_1);
      //linExt(elems,ucs);
      OrderedSet poset;
      try {
        poset = new OrderedSet("Test-Main", elems,ucs);
      } catch (NonOrderedSetException e) {
        System.out.println(e.getMessage());
        return;
      }
      System.out.println("\nThe size is " + poset.card());
    } else {
      OrderedSet test;
      try {
        test = new OrderedSet(new InputLattice(args[0]));
      } catch (NonOrderedSetException e) {
        System.out.println(e.getMessage());
        return;
      }
      ChainDecomposition chainDec = new ChainDecomposition(test);
      Enumeration e = test.univ().elements();
      while (e.hasMoreElements()) {
	System.out.print(" " + ((POElem)e.nextElement()).label());
      }
System.out.println("");
System.out.println("univ: " + test.univ());
System.out.println("chain dec: " + chainDec.numChains());
System.out.println("chain dec: " + chainDec.getHashtable());
System.out.println("chain dec: " + chainDec.getHashtable().size());
System.out.println("elemOrder: " + test.elemOrder);
printElems("Elements: ", test.univ());
System.out.println("testing leq: " + 
test.leq((POElem)test.univ().elementAt(1), (POElem)test.univ().elementAt(2)));
System.out.print("filter of " + ((POElem)test.univ().elementAt(0)).label()
        + " is ");
printElems(" ", ((POElem)test.univ().elementAt(0)).filter());
//System.out.print("highIncomparables of " + 
	//((POElem)test.univ().elementAt(2)).label() + " are ");
//System.out.println(((POElem)test.univ().elementAt(2)).highIncomparables());
for (int i=0; i < test.card(); i++) {
  System.out.print("highIncomparables of " + 
	((POElem)test.univ().elementAt(i)).label() + " are ");
System.out.println(((POElem)test.univ().elementAt(i)).highIncomparables());
}


    }
  }

  static void printElems (String head, Vector v) {
    System.out.println(head);
    Enumeration list = v.elements();
    while (list.hasMoreElements()) {
      System.out.print(" ");
      POElem x = (POElem)list.nextElement();
      System.out.print(x.label());
      System.out.println("\t" + x.rank());
    }
    System.out.println("End");
  }


}

class QuickSortElems extends QuickSort {
  
  public QuickSortElems(Vector v) {
    super(v);
  }

  public boolean leq(Object a, Object b) {
    return ((POElem)a).rank() <= ((POElem)b).rank();
  }

  public boolean lt(Object a, Object b) {
    return ((POElem)a).rank() < ((POElem)b).rank();
  }
}

