package javamath.hasse.util;
/* 
 * LinkedList.java	96/6/26 Ralph Freese
 */

import java.lang.*;
import java.util.*;

/**
 * Linked lists. These are sometimes more efficient that Vector's.
 * For example, the space required to hold all subsets of an 
 * <i>n</i> element set requires space proportional to 
 * (<i>n</i>/2) 2<i><sup>n</sup></i> using vectors, but only
 * </i>2<i><sup>n</sup></i> with linked lists.
 * <p>
 * Each element of the list is held in a Node, which contains a
 * pointer to the next Node or null. The LinkedList object is a 
 * wrapper on top. This means rest() must create a new LinkedList. 
 * But pop and push don't.
 * <p>
 * reverse, nth, nthcdr setnth etc not yet implemented.  
 */
public class LinkedList implements Cloneable {

  protected Node root;

/**
 * Constructs the empty list.
 */
  public LinkedList() {
    root = null;
  }

/**
 * Constructs a list with only obj.
 * @param obj The Object
 */
  public LinkedList(Object obj) {
    root = new Node(obj);
  }

/**
 * Constructs a list with obj followed by list.
 * @param obj The Object to be first.
 * @param list The List to be the rest.
 */
  public LinkedList(Object obj, LinkedList list) {
    root = new Node(obj, list.root);
  }
 
  public boolean isEmpty() {
    return root == null;
  }

/**
 * @exception javamath.hasse.util.EmptyListException The list is empty.
 */
  public Object first() throws EmptyListException {
    if (root == null) throw new EmptyListException();
    return root.contents;
  }

/**
 * This creates a new LinkedList.
 * @exception javamath.hasse.util.EmptyListException The list is empty.
 */
  public LinkedList rest() throws EmptyListException {
    if (root == null) throw new EmptyListException();
    LinkedList list = new LinkedList();
    list.root = this.root.next;
    return list;
  }

/**
 * This modifies the LinkedList.
 */
  public void push(Object obj) {
    Node node = new Node(obj, this.root);
    this.root = node;
  }

/**
 * This modifies the LinkedList.
 * @exception javamath.hasse.util.EmptyListException The list is empty.
 */
  public Object pop() throws EmptyListException {
    Object obj = first();
    this.root = root.next;
    return obj;
  }

  public int size() {
    if (isEmpty()) return 0;
    return root.size();
  }

  public Enumeration elements() {
    return new EnumerationLinkedList(this);
  }

  public Object clone() {
    try {
      return super.clone();
    } catch (CloneNotSupportedException e) {
      // cannot happen
      throw new InternalError(e.toString());
    }
  }

  public String toString() { 
    String str = "("; 
    Enumeration list = this.elements(); 
    while (list.hasMoreElements()) { 
      str += list.nextElement() + " "; 
    } 
    str += ")"; 
    return str; 
  }
   
/*

  public static void main(String[] args) throws EmptyListException {
    LinkedList foo = new LinkedList("first-element");
    System.out.println("first gives " + foo.first());
    System.out.println("rest gives " + foo.rest());
    LinkedList goo = new LinkedList("second-element", foo);
    LinkedList hoo = new LinkedList("third-element", goo);
    System.out.println("rest of goo gives " + goo.rest());
    System.out.println("test foo isEmpty: " + foo.isEmpty());
    System.out.println("test rest of foo isEmpty: " + foo.rest().isEmpty());
    System.out.println("size of goo : " + goo.size());
    System.out.println("test pop on goo : " + goo.pop());
    System.out.println("size of goo : " + goo.size());
    System.out.println("test pop on goo again : " + goo.pop());
    System.out.println("size of goo : " + goo.size());
    System.out.println("size of hoo : " + hoo.size());
    Enumeration e = hoo.elements();
    while(e.hasMoreElements()) {
      System.out.println("the next element is " + e.nextElement()); 
    }
    LinkedList joo = (LinkedList)hoo.clone();
    System.out.println("pop joo : " + joo.pop());
    System.out.println("length of joo : " + joo.size());
    System.out.println("length of hoo : " + hoo.size());
    System.out.println("first of joo : " + joo.first());
    System.out.println("first of hoo : " + hoo.first());
  }

*/

}

class Node {
  Object contents;
  Node next;

  Node(Object obj) {
    contents = obj;
    next = null;
  }
  Node(Object obj, Node node) {
    contents = obj;
    next = node;
  }

  int size() {
    if (next == null) return 1;
    return 1 + next.size();
  }
}

class EnumerationLinkedList implements Enumeration {
  private LinkedList llist;

  EnumerationLinkedList(LinkedList list) {
    llist = new LinkedList();
    llist.root = list.root;
  }

  public boolean hasMoreElements() {
    return (! llist.isEmpty());
  }

  public Object nextElement() {
    if (llist.isEmpty()) {
      return null;
    } else {
      try {
	return llist.pop();
      } catch (EmptyListException e) {
	// can't happen
	throw new InternalError(e.toString());
      }
    }
  }
}
