/*
 * Universal Algebra Calculator
 * Copyright (C) 2001 Ralph Freese (ralph@math.hawaii.edu)
 *                    Emil W. Kiss (ewkiss@cs.elte.hu)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * The project's pages are at
 *     http://www.cs.elte.hu/~ewkiss/software/uaprog
 *     http://www.math.hawaii.edu/~ralph/software/uaprog
 */

package uacalc;

import uacalc.alg.Algebra;
import uacalc.util.SwingWorker;
import uacalc.util.Project;
import uacalc.util.ewkFileChooser;
import uacalc.util.ewkGetNumberDialog;
import uacalc.vectors.VectorList;
import uacalc.vectors.FreeVectorList;
import uacalc.alg.con.CongruenceLattice;
import uacalc.alg.con.ui.CongruenceDisplayerFrame;

import uacalc.ex.AlgebraWriteException;
import uacalc.ex.BadVectorListException;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

/**
 * This class contains the functionality of the Compute menu of the 
 * main frame {@link UACalc} of the Universal Algebra Calculator.
 * It allows the user to create direct products, generate a subpower,
 * compute a free algebra, find all unary polynomials, to interrupt
 * calculations, and to bring up the 
 * {@link uacalc.alg.con.ui.CongruenceDisplayerFrame} which allows
 * working with the congruence lattice of an algebra.
 */
public class Compute {

/**
 * Prepares the <code>Compute</code> menu,
 * and attaches it to the main frame.
 *
 * @param parent The main frame of the calculator.
 * @param menuBar The menu bar of the main frame.
 */
    public Compute(UACalc parent, JMenuBar menuBar) {

	MainFrame = parent;

//Compute Menu
	JMenu ComputeMenu = new JMenu("Compute");
	ComputeMenu.setMnemonic(KeyEvent.VK_C);
	menuBar.add(ComputeMenu);

//Subproducts submenu
	JMenu HSPSubMenu = new JMenu("Subproducts");
	HSPSubMenu.setMnemonic(KeyEvent.VK_U);
	ComputeMenu.add(HSPSubMenu);

//DirectProduct
        JMenuItem DirectProductMenuItem = 
	    new JMenuItem("Direct product", KeyEvent.VK_P);
	HSPSubMenu.add(DirectProductMenuItem);
	DirectProductMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.CTRL_MASK));

	DirectProductMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnDirectProduct();
		}
	    });

//SubPower
        JMenuItem SubPowerMenuItem = 
	    new JMenuItem("Generate subpower", KeyEvent.VK_W);
	HSPSubMenu.add(SubPowerMenuItem);
	SubPowerMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_W, ActionEvent.CTRL_MASK));

	SubPowerMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnSubPower();
		}
	    });

//Congruences
        JMenuItem CongruencesMenuItem = 
	    new JMenuItem("Congruences ...", KeyEvent.VK_C);
	ComputeMenu.add(CongruencesMenuItem);
	CongruencesMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.CTRL_MASK));

	CongruencesMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnCongruences();
		}
	    });

//Terms submenu
	JMenu TermsSubMenu = new JMenu("Terms");
	TermsSubMenu.setMnemonic(KeyEvent.VK_R);
	ComputeMenu.add(TermsSubMenu);

//Unary
        JMenuItem UnaryMenuItem = 
	    new JMenuItem("Unary polynomials", KeyEvent.VK_U);
	TermsSubMenu.add(UnaryMenuItem);
	UnaryMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_U, ActionEvent.CTRL_MASK));

	UnaryMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnUnary();
		}
	    });

//FreeAlg
        JMenuItem FreeAlgMenuItem = 
	    new JMenuItem("Free algebras", KeyEvent.VK_F);
	TermsSubMenu.add(FreeAlgMenuItem);
	FreeAlgMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_F, ActionEvent.CTRL_MASK));

	FreeAlgMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnFreeAlg();
		}
	    });

//Stop
        JMenuItem StopMenuItem = new JMenuItem("Stop", KeyEvent.VK_S);
	ComputeMenu.add(StopMenuItem);
	StopMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.CTRL_MASK));

	StopMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnStop();
		}
	    });

    }


/**
 * Callback function for the Compute/Subproducts/Direct product menu.
 * This is a wrapper function, which calls {@link #ComputeDirectProduct},
 * and makes sure of locking (see {@link Progress#ProgramStart}).
 */
    private void OnDirectProduct() {
	if( !Progress.ProgramStart() ) return;
	if( !ComputeDirectProduct() ) {
	    Progress.ProgramFinished();
	}
    }

/**
 * Compute the direct product of algebras.
 * The computation is done in java, and no separate thread is started.
 * @return true if the computation started, but is not yet finished.
 *
 * Since this function does not use a thread to compute,
 * it always returns false.
 */
    private boolean ComputeDirectProduct() {

	ewkGetNumberDialog dialog = new ewkGetNumberDialog(
	    MainFrame,
	    "Creating a direct product of n algebras.\n" +
	    "Please, enter the number n:",
	    -1, 1, VectorList.VEC_LEN_LIMIT-1);
	int comp = dialog.getValidatedNumber();
	if(comp == -1) {
	    return false;
	}
	String fn[] = new String[comp];
	ArrayList components = new ArrayList(comp);
	Project fn_Project[] = new Project[comp];
	int[] limits = new int[comp];
	int i;
	for( i=0; i<comp; i++ ) {
	    fn[i] = ewkFileChooser.getFile(
		MainFrame, 
		Project.getAlgebraFilter(),
		"Select algebra #" + (i+1),
		ewkFileChooser.OPEN);
	    if(fn[i] == null) return false;
	    fn_Project[i] = new Project(fn[i]);
	    Algebra A = new Algebra(fn[i]);
	    if( !A.ReadAlgebra() ) return false;
	    if( i > 0 ) {
		String s = ((Algebra)components.get(0)).isSimilar(A);
		if( s != null ) {
		    JOptionPane.showMessageDialog(
			MainFrame,
			s + "\nPlease, select another algebra.",
			"Algebras are not similar",
			JOptionPane.ERROR_MESSAGE);
		    i--;
		    continue;
		}
	    }
	    components.add(A);
	    limits[i] = A.getSize();
	}

	//Prepare the universe of the direct product.
	VectorList universe;
	try {
	    universe = VectorList.directProduct(limits);
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend("\n!! Creating universe unsuccesful:\n" +
			       e.toString() + "\n");
	    return false;
	}

	//get the name of the product
	String message = new String(
	    "is the direct product of the algebras\n");
	for( i=0; i<comp; i++ ) {
	    message += fn[i] + "\n";
	}
	//TEMPORARY:
	message += "The universe is:\n" + universe.toString();

	String prod =
	ewkFileChooser.getFile(MainFrame, 
			       Project.getAlgebraFilter(),
			       "Name of the direct product",
			       ewkFileChooser.NEW);
	if(prod == null) return false;

	for( i=0; i<comp; i++ ) {
	    fn_Project[i].lock(true);
	}

	Project prod_Project = new Project(prod);
	if( !prod_Project.create(
	    MainFrame, "This algebra " + message) ) {
	    for( i=0; i<comp; i++ ) {
		fn_Project[i].lock(false);
	    }
	    return false;
	}
	for( i=0; i<comp; i++ ) {
	    fn_Project[i].lock(false);
	}
	//Name may have changed
	prod = prod_Project.getName(Project.AlgebraType);


	//Save universe of the direct product.
	try {
	    FileOutputStream fout = new FileOutputStream(
		prod_Project.getName(Project.UniverseType));
	    PrintStream fp =
		new PrintStream(fout);
	    universe.Write(fp, Project.UniverseType);
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend("\n!! Saving universe unsuccesful:\n" +
			       e.toString() + "\n");
	    return false;
	}

	//Compute and save the direct product
	try {
	    Algebra P = 
		Algebra.directProduct(components, universe, prod);
	    FileOutputStream fout =
		new FileOutputStream(prod);
	    PrintStream fp =
		new PrintStream(fout);
	    P.Write(fp);
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend("\n!! Creating direct product unsuccesful:\n" +
			       e.toString() + "\n");
	    return false;
	}
	Progress.ResultLogAppend("\n" + prod + "\n" + message + "\n");
	Progress.ProgramFinished();
	return false;
    }

/**
 * Callback function for the Compute/Subproducts/Generate subpower menu.
 * This is a wrapper function, which calls {@link #ComputeSubPower},
 * and makes sure of locking (see {@link Progress#ProgramStart}).
 */
    private void OnSubPower() {
	if( !Progress.ProgramStart() ) return;
	if( !ComputeSubPower() ) {
	    Progress.ProgramFinished();
	}
    }

/**
 * Generate a subpower of an algebra.
 * The computation is done in C, and a separate thread is started
 * via {@link uacalc.util.SwingWorker}.
 * @return true if the computation started, but is not yet finished.
 */
    private boolean ComputeSubPower() {

//Later this should be more complicated, we have to give
//the option to load an existing vectorlist file!!
//At the moment, we input everything.

	ewkGetNumberDialog dialog = new ewkGetNumberDialog(
	    MainFrame,
	    "Generating a subalgebra in the n-th power of an algebra.\n" +
	    "Please, enter the number n:",
	    -1, 1, VectorList.VEC_LEN_LIMIT-1);
	int length = dialog.getValidatedNumber();
	if(length == -1) {
	    return false;
	}
	String fn = 
	    ewkFileChooser.getFile(MainFrame, 
				   Project.getAlgebraFilter(),
				   "Select the algebra",
				   ewkFileChooser.OPEN);
	if(fn == null) return false;

	final Algebra A = new Algebra(fn);
	if( !A.ReadAlgebra() ) return false;
	final int size = A.getSize();
	int[] sizes = new int[length];
	int i;
	for( i=0; i<length; i++) {
	    sizes[i]=size;
	}
	final VectorList list = new VectorList();
	list.setLength(length);
//Read vectors
	if( !list.Input(MainFrame, sizes,
			"Please enter generator") ) return false;

	String gen =
	ewkFileChooser.getFile(MainFrame, 
			       Project.getAlgebraFilter(),
			       "Name of the generated algebra",
			       ewkFileChooser.NEW);
	if(gen == null) return false;

	final Project fn_Project = new Project(fn);
	fn_Project.lock(true);

	final Project gen_Project = new Project(gen);
	String message = 
	    "This algebra is the subalgebra of the " +
	    length + "-th power of the algebra\n" +
	    fn + "\ngenerated by:\n";
	message += list.toString();
	if( !gen_Project.create(
	    MainFrame, message) ) {
	    fn_Project.lock(false);
	    return false;
	}
	fn_Project.lock(false);

	//Name may have changed
	gen = gen_Project.getName(Project.AlgebraType);

	String gen_name = 
	    gen_Project.getName(Project.VectorListType);

	try {
            FileOutputStream fout =
                new FileOutputStream(gen_name);
            PrintStream fp =
                new PrintStream(fout);
            list.Write(fp, Project.VectorListType);
            fp.close();
            fout.close();
        }
        catch(Exception e) {
	    System.out.println(e.toString());
	    return false;
        }

	try {
	    FileOutputStream fout =
		new FileOutputStream(
		    gen_Project.getName(Project.ParType));
	    PrintStream fp =
		new PrintStream(fout);
	    fp.println("\\begin_command");
	    fp.println("\\generate_subalgebra");
	    fp.println("\\algebra_file{" +
		       fn_Project.getName(Project.AlgebraType) + "}");
	    fp.println("\\vectorlist_file{" +
		       gen_name + "}");
	    fp.println("\\output_file{" + 
		       gen_Project.getName(Project.UniverseType) + "}");
	    fp.println("\\end_command");
	    fp.println("%%");
	    fp.println("\\begin_command");
	    fp.println("\\create_subproduct_algebra");
	    fp.println("\\vectorlist_file{" +
		       gen_Project.getName(Project.UniverseType) + "}");
	    for( i=0; i < list.getLength(); i++) {
		fp.println("\\algebra_file{" +
			   fn_Project.getName(Project.AlgebraType) + "}");
	    }
	    fp.println("\\output_file{" + 
		       gen_Project.getName(Project.AlgebraType) + "}");
	    fp.println("\\end_command");
	    fp.close();
	    fout.close();
	    if( fp.checkError() ) {
		throw new AlgebraWriteException(//!!!!!!!!!!!!!!!!
		    "Error writing parameter file\n" +
		    gen_Project.getName(Project.ParType) + "\n");
	    }
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend(e.toString());	    
	}

	worker = new SwingWorker() {
		public Object construct() {
		    run_uab(gen_Project.getName(Project.NoType), defaultLogLevel);
		    return "";
		}
		public void finished() {
		    Progress.ProgramFinished();
		    if( !gen_Project.existsName(Project.UniverseType) ) {
			return;//Error message??
		    }
		    VectorList universe = new VectorList();
		    if( !universe.ReadVectorList(
			gen_Project.getName(Project.UniverseType),
			size, Project.UniverseType )) {
			return;
		    }
		    gen_Project.appendToLog(
			"\n > The generated universe is:\n" +
			universe.toString());
		    //!! Should we do this if the universe is HUGE??
		    //Later, we shall have displayers that show things
		    //like congruences, minsets, in terms of the NAME
		    //of the elements of the algebra. A NAME can be
		    //a tuple if the algebra is a subpower.
		}

	    };

//Write to result places
	String res = 
	    "\n [ Generating algebra " + gen + ".\n" +
	    "   Generator(s): \n" + list.toString();
	Progress.ResultLogAppend(res);
	worker.start();  //required for SwingWorker 3
	return true;
    }



/**
 * Callback function for the Compute/Congruences menu.
 * Inputs an algebra name, and starts 
 * {@link uacalc.alg.con.ui.CongruenceDisplayerFrame} 
 * for this algebra.
 */
    public void OnCongruences() {
	String fn = ewkFileChooser.getFile(
	    MainFrame, 
	    Project.getAlgebraFilter(),
	    "Select algebra",
	    ewkFileChooser.OPEN);
	if(fn == null) return;

	Algebra A = new Algebra(fn);
	if( !A.ReadAlgebra() ) return;

	CongruenceLattice congs = new CongruenceLattice(A);
	congs.importProject();
	CongruenceDisplayerFrame displayer = 
	    new CongruenceDisplayerFrame(MainFrame, congs);
	//!!Disable this window?????

    }

/**
 * Callback function for the Compute/terms/Unary menu.
 * This is a wrapper function, which calls {@link #ComputeUnary},
 * and makes sure of locking (see {@link Progress#ProgramStart}).
 */
    private void OnUnary() {
	if( !Progress.ProgramStart() ) return;
	if( !ComputeUnary() ) {
	    Progress.ProgramFinished();
	}
    }

/**
 * Compute all unary polynomials of an algebra.
 * The computation is done in C, and a separate thread is started
 * via {@link uacalc.util.SwingWorker}.
 * @return true if the computation started, but is not yet finished.
 */
    private boolean ComputeUnary() {

	String fn = ewkFileChooser.getFile(
	    MainFrame, 
	    Project.getAlgebraFilter(),
	    "Select the algebra to compute its unary polynomials",
	    ewkFileChooser.OPEN);
	if(fn == null) return false;

	final Project fn_Project = new Project(fn);

	boolean existsP1 = 
	    fn_Project.existsName(Project.UnaryPolType) &&
	    fn_Project.existsName(Project.IdempotentUnaryPolType);

	if( existsP1 ) {
	    Progress.ProgressLogAppend("!! Calculation has already been performed.");
	    return false;
	}

	try {
	    FileOutputStream fout =
		new FileOutputStream(
		    fn_Project.getName(Project.ParType));
	    PrintStream fp =
		new PrintStream(fout);
	    fp.println("\\begin_command");
	    fp.println("\\unary_polynomials");
	    fp.println("\\algebra_file{" +
		       fn_Project.getName(Project.AlgebraType) + "}");
	    fp.println("\\output_file{" + 
		       fn_Project.getName(Project.UnaryPolType) + "}");
	    fp.println("\\output_file{" + 
		       fn_Project.getName(Project.IdempotentUnaryPolType) + "}");
	    fp.println("\\end_command");
	    fp.close();
	    fout.close();
	    if( fp.checkError() ) {
		throw new AlgebraWriteException(//!!!!!!!!!!!!!!!!
		    "Error writing parameter file\n" +
		    fn_Project.getName(Project.ParType) + "\n");
	    }
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend(e.toString());	    
	}

	worker = new SwingWorker() {
		public Object construct() {
		    run_uab(fn_Project.getName(Project.NoType), defaultLogLevel);
		    return "";
		}
		public void finished() {
		    Progress.ProgramFinished();
		}

	    };
	Progress.ResultLogAppend("\n [ Unary polynomials of " + fn);
	worker.start();  //required for SwingWorker 3
	return true;
    }


/**
 * Callback function for the Compute/terms/Free algebras menu.
 * This is a wrapper function, which calls {@link #ComputeFreeAlg},
 * and makes sure of locking (see {@link Progress#ProgramStart}).
 */
    private void OnFreeAlg() {
	if( !Progress.ProgramStart() ) return;
	if( !ComputeFreeAlg() ) {
	    Progress.ProgramFinished();
	}
    }

/**
 * Compute a free algebra in a finitely generated variety.
 * The computation is done in C, and a separate thread is started
 * via {@link uacalc.util.SwingWorker}.
 * @return true if the computation started, but is not yet finished.
 */
    private boolean ComputeFreeAlg() {

	ewkGetNumberDialog dialog = new ewkGetNumberDialog(
	    MainFrame,
	    "Generating a free algebra on n generators.\n" +
	    "Please, enter the number n:",
	    -1, 0, VectorList.VEC_LEN_LIMIT-1);
	int rank = dialog.getValidatedNumber();
	if(rank == -1) {
	    return false;
	}
	String fn = 
	    ewkFileChooser.getFile(MainFrame, 
				   Project.getAlgebraFilter(),
				   "Select the generating algebra of the variety",
				   ewkFileChooser.OPEN);
	if(fn == null) return false;

	Algebra A = new Algebra(fn);
	if( !A.ReadAlgebra() ) return false;
	int size = A.getSize();
	FreeVectorList list;
	try {
	    list = new FreeVectorList(size, rank);
	}
	catch( BadVectorListException e) {
	    Progress.ProgressLogAppend(e.toString());	    
	    return false;
	}

	String gen =
	ewkFileChooser.getFile(MainFrame, 
			       Project.getAlgebraFilter(),
			       "Name of the generated free algebra",
			       ewkFileChooser.NEW);
	if(gen == null) return false;

	final Project fn_Project = new Project(fn);
	fn_Project.lock(true);

	final Project gen_Project = new Project(gen);
	String message = 
	    "This is the free algebra on " + rank + 
	    " generators in the variety generated by the algebra\n" +
	    fn + ".\n";
	if( !gen_Project.create(
	    MainFrame, message) ) {
	    fn_Project.lock(false);
	    return false;
	}
	fn_Project.lock(false);

	//Name may have changed
	gen = gen_Project.getName(Project.AlgebraType);

	String gen_name = 
	    gen_Project.getName(Project.VectorListType);

	try {
            FileOutputStream fout =
                new FileOutputStream(gen_name);
            PrintStream fp =
                new PrintStream(fout);
            list.Write(fp, Project.VectorListType);
            fp.close();
            fout.close();
        }
        catch(Exception e) {
	    System.out.println(e.toString());
	    return false;
        }

	try {
	    FileOutputStream fout =
		new FileOutputStream(
		    gen_Project.getName(Project.ParType));
	    PrintStream fp =
		new PrintStream(fout);
	    fp.println("\\begin_command");
	    fp.println("\\generate_subalgebra");
	    fp.println("\\algebra_file{" +
		       fn_Project.getName(Project.AlgebraType) + "}");
	    fp.println("\\vectorlist_file{" +
		       gen_name + "}");
	    fp.println("\\output_file{" + 
		       gen_Project.getName(Project.UniverseType) + "}");
	    fp.println("\\end_command");
	    fp.println("%%");
	    fp.println("\\begin_command");
	    fp.println("\\create_subproduct_algebra");
	    fp.println("\\vectorlist_file{" +
		       gen_Project.getName(Project.UniverseType) + "}");
	    int i;
	    for( i=0; i < list.getLength(); i++) {
		fp.println("\\algebra_file{" +
			   fn_Project.getName(Project.AlgebraType) + "}");
	    }
	    fp.println("\\output_file{" + 
		       gen_Project.getName(Project.AlgebraType) + "}");
	    fp.println("\\end_command");
	    fp.close();
	    fout.close();
	    if( fp.checkError() ) {
		throw new AlgebraWriteException(//!!!!!!!!!!!!!!!!
		    "Error writing parameter file\n" +
		    gen_Project.getName(Project.ParType) + "\n");
	    }
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend(e.toString());	    
	}

	worker = new SwingWorker() {
		public Object construct() {
		    run_uab(gen_Project.getName(Project.NoType), defaultLogLevel);
		    return "";
		}
		public void finished() {
		    Progress.ProgramFinished();
		}
	    };

//Write to result places
	String res = 
	    "\n [ Generating the " + rank + "-generated free algebra "
	    + gen + "\n" +
	    "   in the variety generated by " + fn + ".\n";
	Progress.ResultLogAppend(res);
	worker.start();  //required for SwingWorker 3
	return true;
    }



/**
 * Callback function for the Compute/Stop menu.
 * Tries to stop the ongoing calculation.
 * Works for the C part (see {@link #stop_uab}), and for the
 * java part (see {@link Progress#StopProgram}).
 */
    public void OnStop() {
	stop_uab();
	Progress.StopProgram = true;
	Progress.ProgramFinished();
    }

/**
 *  The Swing worker used for the calculations.
 */
    private SwingWorker worker;
/**
 *  The frame containing the compute menu.
 */
    private UACalc MainFrame;
/**
 *  The default log_level (see {@link #run_uab}).
 */
    public static final int defaultLogLevel = 2;

//////////////////////////////////////////
//Native stuff

/**
 *  Runs a calculation using the C part.
 *  It does so by preparing a "parameter file" <code>project.par</code>
 *  (see {@link uacalc.util.Project}),
 *  and hands it to the batch version of the program.
 *
 *  @param project The name of the project to run.
 *  @param log_level Verbosity level for the log file written by the C part.
 */
    public static native void run_uab(String project, int log_level);

/**
 *  Try to interrupt the calculation done by the C part.
 *  @see #OnStop
 */
    public native void stop_uab();

/**
 *  Load the dynamic library containing the part written in C.
 */
    static {
	System.loadLibrary("ua");
    }
}

