package uacalc.alg.con;

import uacalc.vectors.IntArray;

/* BasicPartition.java 2001/06/04 Ralph Freese */
/* Added getNumberOfBlocks(),
 * new method Kernel()
 * implemented partToBlockString()
 * added zero() and one()
 * new default constructor
 * Added converter toMatt();
 *   by Emil W. Kiss 2001/06/11
 * tried to make sure that this is always normalized
 */

import java.util.*;
import java.io.*;

/**
 * This class implement the basic operations for partition on the set
 * {0, 1, ..., n-1}. A class that wants to implement partitions on an
 * arbibtrary set can use this class internally. It can also be used
 * as a wrapper for an array of int's in order to use the equals and
 * hashCode methods.
 * <p>
 * It is based on my unpublished note: <i>Partition Algorithms</i>
 * which can be obtained at 
 * {@link <a http://www.math.hawaii.edu/~ralph/Notes/}.
 *
 *
 *
 * @author Ralph Freese
 * @version $Id:$
 */
public class BasicPartition extends IntArray implements Partition {

  /** EWK
   */
// rsf we don't need this but if we did I would make it represent
//     a partition on the empty set. So I would set array = new int[0]
//     and size = 0;
//
//  public BasicPartition() {
//    array = null;
//    size = -1;
//  }

  public BasicPartition(int[] part) {
    this.array = part;
    this.size = part.length;
    normalize();
  }

// rsf  make these static methods

  public static BasicPartition zero(int asize) {
    int[] array = new int[asize];
    for(int i=0; i < asize; i++ ) {
      array[i]= -1;
    }
    return new BasicPartition(array);
  }

  public static BasicPartition one(int asize) {
    int[] array = new int[asize];
    if (asize > 0) {       // in case anyone wants to make a 0 elem alg
      array[0] = -asize;
      for(int i=1; i < asize; i++ ) {
        array[i]= 0;
      }
    }
    return new BasicPartition(array);
  }


    public int getSize() {
	return size;
    }

    public boolean isRelated(int i, int j) {
	if(representative(i) == representative(j) ) {
	    return true;
	}
	return false;
    }


    /** EWK
     * Does not need normalized form.
     * Simply counts the negative entries.
     */
    public int getNumberOfBlocks() {
	int count = 0;
	for( int i=0; i<array.length; i++ ) {
	    if( array[i]<0) count++;
	}
	return(count);
    }

    /** EWK
     * this.array can be ANY array of integers.
     * This function takes its kernel, and prepares
     * the corresponding partition in normalized form.
     */
// rsf: changing this to a static method
    public static BasicPartition Kernel( int[] part ) {

	int size = part.length;
	int[] array = new int[size];
	HashMap inverse = new HashMap(size);
	int i;
	Integer v;
	int j;
	for( i=0; i<size; i++ ) {
	    v= new Integer(part[i]);
	    if( inverse.containsKey(v) ) {
		j = ((Integer)inverse.get(v)).intValue();
		array[i] = j;
		array[j]--; //j is the root of i
	    } else {
		array[i] = -1;
		inverse.put(v, new Integer(i) );
	    }
	}
        return new BasicPartition(array);
    }

    /** EWK
     * Convert to Matt's original format:
     * f(i)=the "number" of the block if i, where
     * blocks are numbered starting from zero.
     * array is assumed to be in normalized form.
     */
    public int[] toMatt() {

	int[] matt = new int[size];
	int i;
	int c=0;
	for(i=0; i<size; i++) {
	    if(array[i]<0 ) {
		matt[i]=c;
		c++;
	    } else {
		matt[i]=matt[array[i]];
	    }
	}
	return matt;
    }

  /**
   * This finds the root of the tree containing i and also modifies
   * the partition.
   */
  public int root(int i) {
    return root(i, array);
  }

  /**
   * This finds the root of the tree containing i and also modifies
   * the partition.
   */
  public static int root(int i, int[] part) {
    int j = part[i];
    if (j < 0) return i;
    int r = root(j, part);
    part[i] = r;
    return r;
  }

  /** 
   * Note r and s must be roots and distinct.
   */
  public void joinBlocks(int r, int s) {
    joinBlocks(r, s, array);
  }

  /** 
   * Note r and s must be roots and distinct.
   */
  public static void joinBlocks(int r, int s, int[] part) {
    int sizeR = - part[r];
    int sizeS = - part[s];
    if (sizeR < sizeS) {
      part[r] = s;
      part[s] = -(sizeR + sizeS);
    } 
    else {
      part[s] = r;
      part[r] = -(sizeR + sizeS);
    } 
  }

  public Partition join(Partition part2) {
    return join(this.array, part2.toArray());
  }

  public static BasicPartition join(int[] u, int[] v) {
    int n = u.length;
    int[] ans = new int[n];
    for (int i = 0; i < n; i++) {
      ans[i] = v[i];
    }
    for (int i = 0; i < n; i++) {
      if (u[i] >= 0) {
        int r = root(i, ans);
        int s = root(u[i], ans);
        if (r != s) joinBlocks(r, s, ans);
      }
    }
    return new BasicPartition(ans);
  }


  public Partition meet(Partition part2) {
    return meet(this.array, part2.toArray());
  }

  public static BasicPartition meet(int[] u, int[] v) {
    int n = u.length;
    HashMap ht = new HashMap();
    int[] ans = new int[n];
    for (int i = 0; i < n; i++) {
      // Get the roots and make a "partition" [r1, r2] so we can 
      // store them on a HT.
      IntArray rootPair 
               = new IntArray(new int[] {root(i, u), root(i, v)});
      Integer rootInt = (Integer)ht.get(rootPair);
      if (rootInt != null) {
        int r = rootInt.intValue();
        ans[r]--; // increase the size by 1 (since it is stored as a neg int)
        ans[i] = r;
      }
      else {
        ht.put(rootPair, new Integer(i));
        ans[i] = -1;
      }
    }
    return new BasicPartition(ans);
  }

  public boolean leq(Partition part2) {
    return leq(this.array, part2.toArray());
  }

  public static boolean leq(int[] u, int[] v) {
    int n = u.length;
    for (int i = 0; i < n; i++) {
      if (u[i] >= 0 && (root(i, v) != root(root(i, u), v))) return false;
    }
    return true;
  }
  
  /**
   * This modifies this.array.
   */
  public void normalize() {
    normalize(array);
  }

  /** 
   * Modify <code>part</code> so that it is in normal form.
   */
  public static void normalize(int[] part) {
    int size = part.length;
    for (int i = 0; i < size; i++) {
     try {
     if (part[0] == 0) throw new Exception();
      int r = root(i, part);
      if (r > i) {
        part[i] = part[r];
        part[r] = i;
      }
     } catch (Exception e) { e.printStackTrace(); }
    }
    for (int i = 0; i < size; i++) {
      root(i, part);
    }
  }

  public String toString(int kind) {
    switch (kind) {
      case INTERNAL:
        return intArrayToString(array);
      case EWK:
        return partToKissString(array);
      case BLOCK:
        return partToBlockString(array);
      case HUMAN:
        return partToBlockString(array) +
	    " (" + getNumberOfBlocks() + " block(s))";
    }
    return intArrayToString(array);
  }

  /**
   * Make String representation of the partition for 
   * the .con and related files. 
   */
  public static String partToKissString(int [] part) {
    StringBuffer sb = new StringBuffer(",");
    for (int i = 0; i < part.length; i++) {
      sb.append(String.valueOf(root(i, part)));
      sb.append(",");
    }
    sb.setLength(sb.length() - 1);
    return sb.toString();
  }


  /** EWK
   * Make String representation of the partition in block form.
   */
  private static String partToBlockString(int [] part) {
    //It is assumed that part is normalized.
    int size = part.length;
    boolean isZero = true;
    for(int i = 0; i < size; i++) {
      if (part[i] != -1) {
        isZero = false;
        break;
      }
    }
    if( isZero ) {
      return("the \"zero\" partition");
    }
    if( part[0] == -part.length ) {
      return("the \"one\" partition");
    }

    ArrayList blocks[] = new ArrayList[part.length];
    int i;
    for(i=0; i<part.length; i++) {
      int r = root(i,part);
      if (blocks[r] == null) blocks[r] = new ArrayList();
      blocks[r].add(new Integer(i));
    }
    String vert = "|";
    String dash = "-";
    StringBuffer sb = new StringBuffer(vert);
    boolean first;
    for(i=0; i<part.length; i++) {
      if( blocks[i] == null ) continue;
      Iterator it = blocks[i].iterator();
      first = true;
      while( it.hasNext() ) {
        if ( !first ) {
          sb.append(dash);
        } else {
          first = false;
        }
        sb.append(it.next().toString());
      }
      sb.append(vert);
    }
    return(sb.toString());
  }
  
  /**
   * This is the public way of finding the root. Unlike <code>root</code
   * it does not modify array.
   */
  public int representative(int i) {
    int j = array[i];
    if (j < 0) return i;
    return representative(j);
  }

  public boolean isRepresentative(int i) {
    return representative(i) == i;
  }

  public boolean isRoot(int i) {
    return representative(i) == i;
  }

  public int[] getRepresentatives() {
    int c = 0;
    for (int i = 0; i < size; i++) {
      if (isRoot(i)) c++;
    }
    int[] ans = new int[c];
    c = 0;
    for (int i = 0; i < size; i++) {
      if (isRoot(i)) ans[c++] = i;
    }
    return ans;
  }


}

