package uacalc.alg.con;

import uacalc.alg.Algebra;
import uacalc.alg.Operation;
import uacalc.vectors.VectorList;
import uacalc.vectors.IntArray;
import uacalc.util.Project;
import uacalc.Progress;
import uacalc.util.ewkInputStream;

import uacalc.ex.ComputationStoppedException;

import java.util.*;
import java.io.*;
import javax.swing.*;

/* CongruenceLattice.java 2001/06/04 Ralph Freese */
/* 
 * Changes by EWK: 2001/06/14
 * The computing part is now in a separate class
 * <code>ComputeCongruenceLattice</code>
 * This part is only the I/O.
 * Methods to compute congruence numbers
 * (see upperCoverNumbers, congruenceLookupMap,
 * makeCongruenceLookup()).
 * Routines to export the list of covers and the list
 * of congruences in VectorList format.
 * (see getCovers())
 * Change in the code of makePrincipals();
 * Ability to read the info from a file
 * rather than computing it.
 * We use three files:
 * A file containing the congruences
 * A file containing the principal congruences map;
 * A file containing the covers.
 * These are all vectorlist files formally,
 * and therefore the I/O is not done here,
 * rather, we simply import three vectorlist files.
 * Facilities included for TCTtype labeling
 */

/**
 * The data in this CongruenceLattice can be set in two ways.
 * Either we let it be computed from the algebra,
 * or it can be imported (if it has been computed already).
 * The query and the export operations can only be used
 * after this.
 */

/**
 * @author Ralph Freese
 * @version $Id:$
 */ 
public class CongruenceLattice extends ComputeCongruenceLattice {

  private Operation joinOp = null;
  private Operation meetOp = null;
  private int zeroCongNum = -1;
  private int oneCongNum = -1;



//Helpers:
/**
 * assigns to each partition its congruence number,
 * that is, the "inverse" of <code>congruences</code>
 */
  private HashMap congruenceLookupMap = null;
  private HashMap principalCongruenceNumbers = null;
  private int NumberOfNonzeroPrincipals = -1;
  private HashMap TCTtypeNumbers = null;
  private HashSet typeSet = null;

/**
 * the i-th member of upperCoverNumbers is the int[] array of the
 * numbers of the covers of Ci.
 */
  private Object[] upperCoverNumbers = null;
  private Object[] lowerCoverNumbers = null;
  private int[] meetIrredNumbers = null;
  private int[] joinIrredNumbers = null;

//////////////////////////////
//Constructor:

    public CongruenceLattice(Algebra alg) {
	super(alg);
    }

//////////////////////////////
//INQUIRY

    /**
     * For the query functions to work, we need
     * congruences, upperCoverNumbers, lowerCoverNumbers,
     * principalCongruenceNumbers,
     * meetIrredNumbers, joinIrredNumbers,
     * zeroCongNum, oneCongNum
     * and optionally
     * joinOp, meetOp, TCTtypeNumbers.
     *
     * If computing time or memory is at premium, these
     * can be prepared individually using the functions below.
     * However, the recommended way is to prepare them
     * in one of the two following bulk ways:
     *
     * If data is IMPORTED from files, then call
     * <code>importAll()</code>
     *
     * If data is COMPUTED from the algebra then call
     * <code>computeAll()</code>
     *
     * These functions do not prepare joinOp and MeetOp.
     *
     * Note: if you do not need the principal congruences,
     * then importPrincipals() and makePrincipalCongruenceNumbers()
     * can be safely omitted.
     */

    public void computeAll() {
	setToEmpty();

	Project m_Project = new Project(getAlgebra().getName());
	makePrincipals();
	m_Project.appendToLog(
	    "\n > There are " + congruences.size()
	    + " nonzero principal congruences.");
	makeJICongruences();
	m_Project.appendToLog(
	    "\n > There are " + joinIrredCongruences.size()
	    + " join irreducible congruences.");
	
	try {
	    makeCongruences();
	} catch(ComputationStoppedException e) {
	    Progress.ProgressLogAppend(
	    "\n ! Calculation interrupted.");
	    setToEmpty();
	    return;
	}
	m_Project.appendToLog(
	    "\n > There are " + congruences.size()
	    + " congruences.");

	makeUpperCovers();

	makeCongruenceLookup();
	makePrincipalCongruenceNumbers();
	makeCoverNumbers();
	makeIrredNumbers();
    }

    public void Complete() {
	//This prepares the congruence lattice,
	//AND the TCT Labeling if necessary,
	//and exports if necessary.

	if( isTCTlabeled() ) return;

	//We do not bother trying to import the TCT labeling,
	//because we assume TEMPORARILY that it is always
	//imported (when available) when the lattice is imported.
	//TEMPORARILY we cannot compute labeling either without
	//redoing the lattice. So we import/compute everything now.
	    
	importProject();
	if( isTCTlabeled() ) {
	    Progress.ProgressLogAppend(
		"\n Succesfully imported the labeled congruence lattice of\n"
		+ getAlgebra().getName());
	    return;
	}
	//unsuccesful import, compute
	Progress.ResultLogAppend(
	    "\n [ Computing the labeled congruence lattice of\n" +
	    getAlgebra().getName());

	makeTCTlabeling(); //TEMPORARILY THIS MAKES Congruences, too!
	if( !isTCTlabeled() ) return;
	
	Project m_Project = new Project(getAlgebra().getName());
	m_Project.appendToLog(
	    "\n > The largest subdirectly irreducible factor has " +
	    getMaxSIfactorSize() + " elements.");
	m_Project.appendToLog(
	    "\n > The TCT type set is " +
	    typeSet.toString() + ".");
	exportProject();
	exportTCTlabeling();
    }

    public void importAll(
	VectorList conlist,
	VectorList prinlist,
	VectorList covers,
	boolean labels ) {

	importCongruences(conlist);
	importPrincipals(prinlist);
	importCovers(covers, labels);
	makeCongruenceLookup();
	makeIrredNumbers();
    }

    private void setToEmpty() {
	congruences = null;
	joinOp = null;
	meetOp = null;
	zeroCongNum = -1;
	oneCongNum = -1;
	congruenceLookupMap = null;
	principalCongruenceNumbers = null;
	NumberOfNonzeroPrincipals = -1;
	TCTtypeNumbers = null;
	upperCoverNumbers = null;
	lowerCoverNumbers = null;
	meetIrredNumbers = null;
	joinIrredNumbers = null;
    }


    public void importProject() {
	//!!This should be refined later:
	//we should display partial info.


	Project m_Project = new Project(getAlgebra().getName());

	setToEmpty();
	if( !m_Project.existsName(Project.ConListType) ) {
	    setToEmpty();
	    return;
	}
	VectorList conlist = new VectorList();
	if( !conlist.ReadVectorList(
	    m_Project.getName(Project.ConListType),
	    algSize, Project.ConListType )) {
	    setToEmpty();
	    return;
	}
	if( !m_Project.existsName(Project.PrincipalMapType) ) {
	    setToEmpty();
	    return;
	}
	VectorList prinlist = new VectorList();
	if( !prinlist.ReadVectorList(
	    m_Project.getName(Project.PrincipalMapType),
	    -1, Project.PrincipalMapType )) {
	    setToEmpty();
	    return;
	}
	boolean labels = true;
	VectorList types = new VectorList();
	if( m_Project.existsName(Project.TypeListType) ) {
	    if( !types.ReadVectorList(
		m_Project.getName(Project.TypeListType),
		-1, Project.TypeListType )) {
		setToEmpty();
		return;
	    }
	} else {
	    if( !m_Project.existsName(Project.CoverListType) ) {
		setToEmpty();
		return;
	    }
	    labels = false;
	    if( !types.ReadVectorList(
		m_Project.getName(Project.CoverListType),
		-1, Project.CoverListType )) {
		setToEmpty();
		return;
	    }
	}
	importAll(conlist, prinlist, types, labels );
	return;
    }


    public void exportProject() {
	VectorList congs = exportCongruencesMatt();
	VectorList principals = exportPrincipals();
	VectorList covers = exportCovers(false);
	
	Project m_Project = new Project(getAlgebra().getName());
	FileOutputStream fout;
	PrintStream fp;
	try {
	    fout = new FileOutputStream(
		m_Project.getName(Project.ConListType));
	    fp = new PrintStream(fout);
	    congs.Write(fp, Project.ConListType);
	    fp.close();
	    fout.close();
	    
	    fout = new FileOutputStream(
		m_Project.getName(Project.PrincipalMapType));
	    fp = new PrintStream(fout);
	    principals.Write(fp, Project.PrincipalMapType);
	    fp.close();
	    fout.close();
	    
	    fout = new FileOutputStream(
		m_Project.getName(Project.CoverListType));
	    fp = new PrintStream(fout);
	    covers.Write(fp, Project.CoverListType);
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend(e.toString());	    
	}
	
    }

    public void exportTCTlabeling() {
	VectorList labeling = exportCovers(true);
	
	Project m_Project = new Project(getAlgebra().getName());
	FileOutputStream fout;
	PrintStream fp;
	try {
	    fout = new FileOutputStream(
		m_Project.getName(Project.TypeListType));
	    fp = new PrintStream(fout);
	    labeling.Write(fp, Project.TypeListType);
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    Progress.ProgressLogAppend(e.toString());	    
	}
	
    }

    /**
     * The algebra associated with this congruence lattice.
     */
    public Algebra getAlgebra() {
	return alg;
    }

//INQUIRY: congruences

    /**
     * The cardinality of this congruence lattice.
     */
    public int getSize() {
	if( congruences == null ) return -1;
	return congruences.size();
    }

    /**
     * Returns congruence Cnum.
     */
    public Partition getPartition(int num) {
	if( congruences == null ) return null;
	return (Partition)congruences.get(num);
    }

    /**
     * Returns congruence iterator.
     */
    public Iterator getCongruenceIterator() {
	//I don't know whether this is preferable to
	//let the user have congruences itself.

	if( congruences == null ) return null;
	return congruences.iterator();
    }

    public int getZeroCongruence() {
	return zeroCongNum;
    }

    public int getOneCongruence() {
	return oneCongNum;
    }


//INQUIRY: lattice operations

    public boolean isLEQ(int c1, int c2) {
	BasicPartition p1 = (BasicPartition)congruences.get(c1);
	BasicPartition p2 = (BasicPartition)congruences.get(c2);
	return p1.leq(p2);
    }

    public int getJoin(int c1, int c2) {

	if(joinOp != null) {//Already computed
	    return joinOp.getValue(c1, c2);
	}

	BasicPartition p1 = (BasicPartition)congruences.get(c1);
	BasicPartition p2 = (BasicPartition)congruences.get(c2);
	int value = ((Integer)congruenceLookupMap.get(
		p1.join(p2))).intValue();
	return value;
    }

    public int getMeet(int c1, int c2) {

	if(meetOp != null) {//Already computed
	    return meetOp.getValue(c1, c2);
	}

	BasicPartition p1 = (BasicPartition)congruences.get(c1);
	BasicPartition p2 = (BasicPartition)congruences.get(c2);
	int value = ((Integer)congruenceLookupMap.get(
		p1.meet(p2))).intValue();
	return value;
    }


//INQUIRY: meet and join-irreducibles, covers

    /**
     * If <code>meetIrred</code> is a meet irreducible congruence,
     * this returns its upper cover. Otherwise it returns -1.
     */
    public int upperCover(int meetIrred) {
	return upperStar(meetIrred);
    }

    /**
     * If <code>meetIrred</code> is a meet irreducible congruence,
     * this returns its upper cover. Otherwise it return <code>null</code>.
     * This is the notation used by Freese and Nation.
     */
    public int upperStar(int meetIrred) {
	if( ((int[])upperCoverNumbers[meetIrred]).length != 1) return -1;
	return ((int[])upperCoverNumbers[meetIrred])[0];
    }

    /**
     * If <code>joinIrred</code> is a join irreducible congruence,
     * this returns its lower cover. Otherwise it returns -1
     */
    public int lowerCover(int joinIrred) {
	return lowerStar(joinIrred);
    }

    /**
     * If <code>joinIrred</code> is a join irreducible congruence,
     * this returns its lower cover. Otherwise it returns -1
     * This is the notation used by Freese and Nation.
     */
    public int lowerStar(int joinIrred) {
	if( ((int[])lowerCoverNumbers[joinIrred]).length != 1) return -1;
	return ((int[])lowerCoverNumbers[joinIrred])[0];
    }

    public int[] getUpperCovers(int c) {
	return (int[])upperCoverNumbers[c];
    }


    public int[] getLowerCovers(int c) {
	return (int[])lowerCoverNumbers[c];
    }

    public boolean isMI(int c) {
	if( getUpperCovers(c).length == 1) return true;
	return false;
    }


    public boolean isJI(int c) {
	if( getLowerCovers(c).length == 1) return true;
	return false;
    }

    public int getNumberOfMIs() {
	return meetIrredNumbers.length;
    }


    public int getNumberOfJIs() {
	return joinIrredNumbers.length;
    }


    public int[] getMINumbers() {
	return meetIrredNumbers;
    }


    public int[] getJINumbers() {
	return joinIrredNumbers;
    }


    public int getMaxSIfactorSize() {
	int i;
	int max = 1;
	int tmp;
	for(i=0; i<meetIrredNumbers.length; i++) {
	    tmp = getPartition(meetIrredNumbers[i]).getNumberOfBlocks();
	    if(tmp > max) max = tmp;
	}
	return max;
    }

    /**
     * This returns the set of principal congruences as a Set of Integers.
     */
    public Set getPrincipalNumbersAsSet() {
      HashSet hs = new HashSet();
      Iterator it = principalCongruenceNumbers.values().iterator();
      while (it.hasNext()) {
        hs.add(it.next());
      }
      return hs;
    }

//INQUIRY: TCT types

    public boolean isTCTlabeled() {
	if( TCTtypeNumbers == null ) return false;
	return true;
    }

    /**
     * Get the TCTtype of a cover a-<b.
     * Returns the type as an integer in [1,5],
     * returns 0 if the type has not yet been computed,
     * returns -1 if this is not a cover.
     * A return value 0 does not guarantee that this is a cover!
     * This function is not for testing whether a-< b.
     */
    public int getTCTtype(int a, int b) {

	if( TCTtypeNumbers == null ) return 0;
	Integer num = (Integer)TCTtypeNumbers.get(
	    new IntArray(new int [] {a, b}));
	if (num == null) return -1;
	return num.intValue();
    }

//INQUIRY: Principal congruences

    /**
     * Get the principal congruence generated by a and b. 
     * <code>principalCongruenceNumbers</code> must have already been set.
     */
    public int Cg(int a, int b) {
	if( a == b) return zeroCongNum;
	if (a > b) {
	    int c = a;
	    a = b;
	    b = c;
	}
	Integer num = (Integer)principalCongruenceNumbers.get(
	    new IntArray(new int [] {a, b}));
	if (num == null) return zeroCongNum;
	return num.intValue();
    }

    public int getNumberOfNonzeroPrincipals() {
	return NumberOfNonzeroPrincipals;
    }


//INQUIRY: Generate

    /**
     * Get the congruence generated by the pairs in <code>gen</code>
     * Needs principalCongruenceNumbers, zeroCongNum,
     * congruenceLookupMap.
     * Uses joinOp if that one exists. If not, then
     * uses principalCongruencesLookup if it exists, otherwise
     * principalCongruenceNumbers and zeroCongNum.
     */
    public int Generate(VectorList gen) {
	if( joinOp != null ) {
	    Iterator it = gen.iterator();
	    int cong;
	    int[] pair;
	    int join = zeroCongNum;
	    while (it.hasNext() ) {
		pair = (int[])it.next();
		cong = Cg(pair[0], pair[1]);
		join = joinOp.getValue(cong, join);
	    }
	    return(join);
	} else {
	    return ((Integer)congruenceLookupMap.get(
		GeneratePartition(gen))).intValue();
	}
    }




////////////////////
//Make Number Tables

    /**
     *  Needs congruences, prepares congruenceLookupMap,
     *  zeroCongNum, oneCongNum.
     */
    public void makeCongruenceLookup() {

	int conLatSize = congruences.size();
	congruenceLookupMap = new HashMap(conLatSize);
	int i = 0;
	Iterator it = congruences.iterator();
	while( it.hasNext() ) {
	    BasicPartition part = (BasicPartition)it.next();
	    int blocks = part.getNumberOfBlocks();
	    if(blocks == 1) oneCongNum = i;
	    if(blocks == algSize ) zeroCongNum = i;
	    congruenceLookupMap.put(part, new Integer(i) );
	    i++;
	}
    }

    /**
     *  Needs principalCongruencesLookup, congruenceLookupMap
     *  prepares principalCongruenceNumbers, NumberOfNonzeroPrincipals.
     */
    public void makePrincipalCongruenceNumbers() {
	principalCongruenceNumbers = new HashMap();
	Iterator it = principalCongruencesLookup.entrySet().iterator();
	while( it.hasNext() ) {
	    Map.Entry entry = (Map.Entry)it.next();
	    BasicPartition part = ((BasicPartition)entry.getValue());
	    principalCongruenceNumbers.put(
		entry.getKey(),
		congruenceLookupMap.get(part));
	}
	makeNumberOfNonzeroPrincipals();
    }


    /**
     * Needs principalCongruenceNumbers,
     * prepares NumberOfNonzeroPrincipals
     */
    private void makeNumberOfNonzeroPrincipals() {
	HashSet hs = new HashSet(
	    principalCongruenceNumbers.values());
	NumberOfNonzeroPrincipals = hs.size();
    }


    /**
     * Needs congruences, upperCoversMap, congruenceLookupMap
     * Prepares upperCoverNumbers, lowerCoverNumbers.
     */
    public void makeCoverNumbers() {

	int conLatSize = congruences.size();
	int j;
	upperCoverNumbers = new Object[conLatSize];
	lowerCoverNumbers = new Object[conLatSize];
	ArrayList[] tmpLowerCoverNumbers = new ArrayList[conLatSize];
	int i;
	for(i=0; i<conLatSize; i++) {
	    tmpLowerCoverNumbers[i] = new ArrayList();
	}
	//Translate upper covers and insert lower cover numbers
	for(i=0; i<conLatSize; i++) {
	    ArrayList covlist = (ArrayList)upperCoversMap.get(
		((BasicPartition)congruences.get(i)));
	    int[] covNumbers = new int[covlist.size()];
	    Iterator cit = covlist.iterator();
	    j = 0;
	    while( cit.hasNext() ) {
		BasicPartition p = (BasicPartition)cit.next();
		covNumbers[j]=
		    ((Integer)congruenceLookupMap.get(p)).intValue();
		tmpLowerCoverNumbers[covNumbers[j]].add(new Integer(i));
		j++;
	    }
	    upperCoverNumbers[i]=covNumbers;
	}
	//Translate tmpLowerCoverNumbers to lowerCoverNumbers
	for(i=0; i<conLatSize; i++) {
	    ArrayList list = tmpLowerCoverNumbers[i];
	    int[] covNumbers = new int[list.size()];
	    lowerCoverNumbers[i] = covNumbers;
	    j=0;
	    Iterator it = list.iterator();
	    while( it.hasNext() ) {
		covNumbers[j] =
		    ((Integer)it.next()).intValue();
		j++;
	    }
	}
    }
    
    /**
     * Needs congruences, lowerCoverNumbers, upperCoverNumbers
     * Prepares meetIrredNumbers, joinIrredNumbers
     */
    public void makeIrredNumbers() {
	//This has to work when we import, so we do
	//not use meetIrredCongruences and joinIrredCongruences

	int countMI = 0;
	int countJI = 0;
	int conLatSize = congruences.size();
	int i;
	for(i=0; i<conLatSize; i++) {
	    if( isJI(i) ) countJI++;
	    if( isMI(i) ) countMI++;
	}
	joinIrredNumbers = new int[countJI];
	meetIrredNumbers = new int[countMI];
	int j = 0;
	int m = 0;
	for(i=0; i<conLatSize; i++) {
	    if( isJI(i) ) {
		joinIrredNumbers[j] = i;
		j++;
	    }
	    if( isMI(i) ) {
		meetIrredNumbers[m] = i;
		m++;
	    }
	}
    }

    /**
     * Needs congruences, upperCoverNumbers 
     * Prepares TCTtypeNumbers
     */
    public void makeTCTlabeling() {
	//I am not sure if this is the right approach.
	//We use the language of congruence numbers,
	//but do all the calculations with partitions.

	//There is a BIG problem here!!!!!!!!!
	//If the congruence lattice is imported, then
	//some data structures (like joinIrredCongruences)
	//is not set! If further computation is done, 
	//for example labeling, then great confusion arises:
	//some parts of the congruence lattice are computed
	//in a possibly incompatible way. This has to be prevented.
	//The problem is caused by the fact that importing
	//and computing uses different data structures.
	//The TEMPORARY solution here is that I simply remake
	//the entire congruence lattice.

	computeAll();	//TEMPORARY HACK!!!!!!!!!!!!!!!!!!!
	if( getSize() == -1 ) return;
	TypeFinder finder = new TypeFinder(this);
	typeSet = finder.findTypeSet();
	HashMap JItypes = finder.getJItypes();
	TCTtypeNumbers = new HashMap();

	int conLatSize = congruences.size();
	int i, j;
	for(i=0; i<conLatSize; i++) {
	    Partition alpha = (Partition)congruences.get(i);
	    int[] covers = ((int[])upperCoverNumbers[i]);
	    for(j=0; j < covers.length; j++) {
		Partition beta = (Partition)congruences.get(covers[j]);
		Partition gamma = findJoinIrred(alpha, beta);
		Integer type = (Integer)JItypes.get(gamma);
		TCTtypeNumbers.put(
		    new IntArray(new int[] {i, covers[j]}), type);
	    }
	}
    }

// rsf can't we make the op without making a join table, ie., the joinOp
//     takes two congruence numbers and returns the number of their 
//     join. For example, subclass Operation to something like
//     CongruenceLatticeOperation which stores the congruence lattice
//     and looks up the partition associate with the congruence numbers,
//     forms their join, and returns the number of this.
//     big space savings. 
    
    /**
     * Needs congruences, congruenceLookupMap
     * Prepares joinOp
     */
    public void makeJoinOp() {
	int size = congruences.size();
	joinOp = new Operation(size, 0);

	int arg1, arg2;
	arg1 = 0;
	Iterator it = congruences.iterator();
	int value;
	while (it.hasNext()) {
	    BasicPartition p1 = (BasicPartition)it.next();
	    arg1++;
	    arg2 = 0;
	    Iterator it2 = congruences.iterator();
	    while (it2.hasNext()) {
		value = ((Integer)congruenceLookupMap.get(
		    p1.join((BasicPartition)it2.next()))).intValue();
		arg2++;
		joinOp.setValue(arg1, arg2, value);
	    }
	}
    }

    /**
     * Needs congruences, congruenceLookupMap
     * Prepares meetOp
     */
    public void makeMeetOp() {
	int size = congruences.size();
	meetOp = new Operation(size, 1);

	int arg1, arg2;
	arg1 = 0;
	Iterator it = congruences.iterator();
	int value;
	while (it.hasNext()) {
	    BasicPartition p1 = (BasicPartition)it.next();
	    arg1++;
	    arg2 = 0;
	    Iterator it2 = congruences.iterator();
	    while (it2.hasNext()) {
		value = ((Integer)congruenceLookupMap.get(
		    p1.meet((BasicPartition)it2.next()))).intValue();
		arg2++;
		meetOp.setValue(arg1, arg2, value);
	    }
	}
    }

////////
//IMPORT

    /**
     * Prepares congruences
     */
    public void importCongruences(VectorList list) {
	//No error checking!
	//We assume that the length is algSize
	//and generally that these are the congruences of alg.
	//They can be arbitrary mappings.
	//We copy list, and import the partitions.
	//We must keep the order, because
	//the other files (covers, types, principals)
	//use the congruence numbers.
	//We do not prepare congruencesHash.

	congruences = new ArrayList(list.size());
	Iterator it = list.iterator();
	int[] funct;
	int i = 0;
	while( it.hasNext() ) {
	    funct = (int[])it.next();
	    BasicPartition part = BasicPartition.Kernel(funct);
            congruences.add(part);
	    i++;
        }
    }

    /**
     * Prepares principalCongruenceNumbers and
     * NumberOfNonzeroPrincipals
     */
    public void importPrincipals(VectorList prin) {
	//The format of each entry is a triple
	//a,b,i means: Cg(a,b) = Ci.
	//No error checking!
	//We assume that the data is correct.
	//Assumes that <code>congruences</code> is valid.

	principalCongruenceNumbers = new HashMap(prin.size());
	Iterator it = prin.iterator();
	int[] triple;
	while( it.hasNext() ) {
	    triple = (int[])it.next();
	    principalCongruenceNumbers.put(
		new IntArray(
		    new int[] {triple[0], triple[1]}), 
		new Integer(triple[2]));
        }
	makeNumberOfNonzeroPrincipals();
    }

    /**
     * Prepares upperCoverNumbers, lowerCoverNumbers
     * and optionally TCTtypeNumbers
     */
    public void importCovers(VectorList covs, boolean labels) {
	//The entries are int[]'s:
	//a,b,..., where a-<b.
	//if labels = true, then the third component is the TCTtype.
	//No error checking!
	//We assume that the data is correct.
	//Assumes that <code>congruences</code> is valid.

	int conLatSize = congruences.size();
	upperCoverNumbers = new Object[conLatSize];
	lowerCoverNumbers = new Object[conLatSize];
	if( labels ) {
	    TCTtypeNumbers = new HashMap(covs.size());
	}

	ArrayList[] tmpUpperCoverNumbers = new ArrayList[conLatSize];
	ArrayList[] tmpLowerCoverNumbers = new ArrayList[conLatSize];

	int i;
	for(i=0; i<conLatSize; i++) {
	    tmpUpperCoverNumbers[i] = new ArrayList();
	    tmpLowerCoverNumbers[i] = new ArrayList();
	}

	Iterator it = covs.iterator();
	int[] next;
	while( it.hasNext() ) {
	    next = (int[])it.next();
	    tmpUpperCoverNumbers[next[0]].add(new Integer(next[1]));
	    tmpLowerCoverNumbers[next[1]].add(new Integer(next[0]));
	    if( labels ) {
	    TCTtypeNumbers.put(
		new IntArray(
		    new int[] {next[0], next[1]}), 
		new Integer(next[2]));
	    }
	}

	int j;
	for(i=0; i<conLatSize; i++) {
	    ArrayList ulist = tmpUpperCoverNumbers[i];
	    ArrayList llist = tmpLowerCoverNumbers[i];
	    int[] uNumbers = new int[ulist.size()];
	    int[] lNumbers = new int[llist.size()];
	    upperCoverNumbers[i] = uNumbers;
	    lowerCoverNumbers[i] = lNumbers;
	    j=0;
	    it = ulist.iterator();
	    while( it.hasNext() ) {
		uNumbers[j]=((Integer)it.next()).intValue();
		j++;
	    }
	    j=0;
	    it = llist.iterator();
	    while( it.hasNext() ) {
		lNumbers[j]=((Integer)it.next()).intValue();
		j++;
	    }
	}
    }

////////
//EXPORT

    /** EWK
     * Export the congruences as a <code>VectorList</code>.
     * This uses Ralph Freese's normalized format.
     */
    public VectorList exportCongruences() {

	VectorList v = new VectorList(congruences);
	v.setLength(algSize);
	return(v);
    }

    /** EWK
     * Export the congruences as a <code>VectorList</code>.
     * This uses Matt's original format so that we can write
     * a *.con file compatible with the old program.
     */
    public VectorList exportCongruencesMatt() {

	VectorList v = new VectorList();
	v.setLength(algSize);
	Iterator it = congruences.iterator();
	while( it.hasNext() ) {
	    BasicPartition part = (BasicPartition)it.next();
	    v.add(part.toMatt());
	}
	return v;
    }

    /** EWK
     * Export the principal congruences as a <code>VectorList</code>.
     * The format is described in importPrincipals()
     * Assumes that principalCongruenceNumbers is already made.
     */
    public VectorList exportPrincipals() {
	return exportHashMap(principalCongruenceNumbers);
    }


    /** EWK
     * Export the list of covers as a <code>VectorList</code>.
     * If labels is true then export the TCTtype labels, too.
     */
    public VectorList exportCovers(boolean labels) {
	if( labels ) 
	    return exportHashMap(TCTtypeNumbers);

	int conLatSize = congruences.size();
	VectorList v = new VectorList();
	v.setLength(2);
	int i, j;
	for(i=0; i<conLatSize; i++) {
	    for(j=0; j< ((int[])upperCoverNumbers[i]).length; j++) {
		int[] next = new int[2];
		next[0]=i;
		next[1]=((int[])upperCoverNumbers[i])[j];
		v.add(next);
	    }
	    i++;
	}
	return v;
    }


  /**
   * Exports the congruence lattice as an algebra.
   */ 
    public void exportLattice(Algebra A) {
	//A should be an empty algebra, with its name set.
	//This is very clumsy!! We could use the name
	//derived from the name of the original algebra m_alg.
	//We'll see.
	
	if(joinOp == null) makeJoinOp();
	if(meetOp == null) makeMeetOp();
	int size = congruences.size();
	A.setSize(size);
	A.addOp(0, joinOp);
	A.addOp(1, meetOp);
    } 


    /** EWK
     * Exports the special type of hashmaps we use for
     * principalCongruenceNumbers and TCTtypeNumbers
     */
    private VectorList exportHashMap(HashMap hm) {

	VectorList v = new VectorList();
	v.setLength(3);

	Iterator it = hm.entrySet().iterator();
	IntArray pair;
	int value;
	while( it.hasNext() ) {
	    Map.Entry entry = (Map.Entry)it.next();
	    pair = (IntArray)entry.getKey();
	    value = ((Integer)entry.getValue()).intValue();
	    int[] triple = new int[3];
	    triple[0]=pair.get(0);
	    triple[1]=pair.get(1);
	    triple[2]=value;
	    v.add(triple);
	}
	return v;
    }

// rsf: additions, I need accessors to the Partitions.

    /**
     * Get the principal congruence generated by a and b as a Partition.
     * Priorities:
     * If <code>PrincipalCongruencesLookup</code> exists, then uses it.
     * If not, but <code>congruences</code> 
     * and <code>PrincipalCongruenceNumbers</code>
     * exist, uses them.
     * Otherwise computes it.
     */
    public Partition CgPartition(int a, int b) {
	if( a == b) return zeroCong;
	if (principalCongruencesLookup == null &&
	    congruences != null && principalCongruenceNumbers!= null ) 
	    return getPartition(Cg(a,b));

	if (a > b) {
	    int c = a;
	    a = b;
	    b = c;
	}
	if (principalCongruencesLookup == null) return makeCg(a, b);

	Partition part = (Partition)principalCongruencesLookup.get(
	    new IntArray(new int [] {a, b}));
	if (part == null) return zeroCong;
	return part;
    }

  /**
   * If <code>joinIrred</code> is a join irreducible congruence,
   * this returns its lower cover. Otherwise it return <code>null</code>.
   * This is the notation used by Freese and Nation.
   */
  public Partition lowerStarPartition(Partition joinIrred) {
    if (lowerCoverOfJIs == null) {
      int ji = ((Integer)congruenceLookupMap.get(joinIrred)).intValue();
      if (ji == -1) return null;
      return getPartition(ji);
    }
    return (Partition)lowerCoverOfJIs.get(joinIrred);
  }


  /**
   * If <code>joinIrred</code> is a join irreducible congruence,
   * this returns its lower cover. Otherwise it return <code>null</code>.
   * This is the notation used by Freese and Nation.
   */
  public int[] getGeneratingPair(Partition joinIrred) {
    int[] lc = lowerStarPartition(joinIrred).toArray();
    int[] ji = joinIrred.toArray();
    int[] ans = new int[2];
    for (int i = 0; i < lc.length; i++) {
      if (ji[i] < 0 && ji[i] != lc[i]) {
        ans[0] = i;
        break;
      }
    }
    for (int i = 0; i < lc.length; i++) {
      if (ji[i] == ans[0] && lc[i] != ans[0]) {
        ans[1] = i;
        break;
      }
    }
    return ans;
  }


    /**
     * Get the congruence generated by the pairs in <code>gen</code>
     * Does not use joinOp even if that one exists.
     * Uses Cgpartiton().
     */
    public Partition GeneratePartition(VectorList gen) {
	Iterator it = gen.iterator();
	Partition cong;
	Partition join = zeroCong;
	int[] pair;
	while (it.hasNext() ) {
	    pair = (int[])it.next();
	    cong = CgPartition(pair[0], pair[1]);
	    join = cong.join(join);
	}
	return(join);
    }

  /**
   * Convert a (congruence) number to a label for the lattices drawer.
   * Note if this is changed, drawerLabelToInt should also be changed.
   */
  public static String intToDrawerLabel(int i) {
    return String.valueOf(i);
  }

  /**
   * Convert a drawer label to the congruence number.
   * Note if this is changed, intToDrawerLabel should also be changed.
   */
  public static int drawerLabelToInt(String label) {
    // emil, you had a NumberFormatter. Were you worried
    //   that there might be strings like "1,234"? If so
    //   put it back.
    return Integer.parseInt(label);
  }

  public void exportHasseDiagram() {

    Project m_Project = new Project(getAlgebra().getName());
    FileOutputStream fout;
    PrintStream fp;
    try {
      fout = new FileOutputStream(m_Project.getName(Project.ConHasseType));
      fp = new PrintStream(fout);
      // fp = System.out;
      final int size = getSize();
      // fp.println("\n(\"Con Lat of "
      //                  + m_Project.getName(Project.NoType) + "\"");
      fp.println("\n(\"" + m_Project.getShortName() + "\"");
      for (int i = 0; i < size; i++) {
        fp.print("  (" + intToDrawerLabel(i) + "  (");
        int[] covers = getUpperCovers(i);
        final int c = covers.length;
        boolean first = true;
        for (int j = 0; j < c; j++) {
          if (first) first = false;
          else fp.print(" ");
          if (isTCTlabeled()) {
            fp.print("(");
            fp.print(intToDrawerLabel(covers[j]));
            fp.print(" ");
            fp.print(String.valueOf(getTCTtype(i, covers[j])));
            fp.print(")");
          }
          else {
            fp.print(intToDrawerLabel(covers[j]));
          }
        }
        fp.println("))");
      }
      fp.println(")");
      fp.flush();
      fp.close();
      if (fout != null) fout.close();
    }
    catch(Exception e) {
      Progress.ProgressLogAppend(e.toString());
    }
  }



///////////////////////////////
//TEST

    public static void main(String[] args) throws Exception {
    if (args.length == 0) {
      System.out.println("give an alg file");
      return;
    }
    Algebra alg = new Algebra("test");
    ewkInputStream fp = new ewkInputStream(new FileInputStream(args[0]));
    alg.Read(fp);
    if (fp != null) fp.close();
    long time = System.currentTimeMillis();
    CongruenceLattice con = new CongruenceLattice(alg);
    con.computeAll();
    time = System.currentTimeMillis() - time;

    if (con.algSize > 2) {
      BasicPartition part = con.makeCg(2, 1); // will fail for 1 element algebra
      System.out.println(part.toString(BasicPartition.BLOCK));
    }


    System.out.println("number of nonzero principals is "
                               + con.getNumberOfNonzeroPrincipals());
    System.out.println("number of congruences is " + con.getSize());

    System.out.println("number of ji congruences is "
                               + con.getNumberOfJIs());
    System.out.println("number of mi congruences is "
                               + con.getNumberOfMIs());
    System.out.println("time to compute con was " + time);
  } 



}

