package uacalc.alg.con;

import uacalc.alg.Algebra;
import uacalc.alg.Operation;
import uacalc.vectors.IntArray;
import uacalc.util.ewkInputStream;

import java.util.*;
import java.io.*;

/* TypeFinder.java  2001/06/14  Ralph Freese */

/**
 * A class for finding TCT types of prime quotients in the congruence
 * lattice and for finding the type set of the algebra.
 * It can also be used to find subtraces.
 * <p>
 * The main part of the calculation is to take elements <i>a</i> 
 * and <i>b</i> of the algebra such that
 * Cg(<i>a</i>, <i>b</i>) is join irreducible, and a congruence
 * alpha above the lower cover of Cg(<i>a</i>, <i>b</i>) but not
 * above Cg(<i>a</i>, <i>b</i>) and find <i>c</i> and <i>d</i>
 * such that Cg(<i>c</i>, <i>d</i>) = Cg(<i>a</i>, <i>b</i>) and
 * {<i>c</i>, <i>d</i>} is a subtrace. Then it finds the type of this
 * pair.
 * <p>
 * The algorithm used is similar to that described in J. Berman, E. W. 
 * Kiss, P. Prohle, and A. Szendrei <i>The set of type of a 
 * finitely generated variety</i>, Discrete Math. 112 (1993), 1-20.
 * Preprint available on the second author's web site:
 * {@link http://www.cs.elte.hu/~ewkiss/}.
 * They define a directed graph G(A) with vertices two elements 
 * subsets of A and edges corresponding to images of unary polynomials.
 * <p>
 * Our algorithm restricts this to only those pairs <i>x</i>, <i>y</i>
 * such that Cg(<i>x</i>, <i>y</i>) = Cg(<i>a</i>, <i>b</i>) and 
 * takes advantage of the fact that this graph is a quasiordered set
 * and that a subtrace corresponds to a minimal element of it. This
 * allows us to find a subtrace with only calculating part of the 
 * graph. But in the worst case analysis this is no better than the
 * original algorithm but in most cases it is much faster.
 * <p>
 * One of the nice things about this is the calculations are really taking
 * place in A/<i>alpha</i> even though we do not explicitly for that 
 * quotient. As noted above <i>alpha</i> can be the lower cover of 
 * Cg(<i>a</i>, <i>b</i>) but it also can be a meet irreducible of small
 * index.
 *
 * @author Ralph Freese
 * @version $Id:$
 */
public class TypeFinder {
  //EXPERIMENTAL! WE DO NOT CARE ABOUT NULLARY OPS

  final boolean showTime = false; // set false to get rid of this

  private Algebra A;
  private int algSize;

  private HashSet visited;
  private boolean hasInvolution = false;

  // this will only have the JI's and their lower covers and lookup
  private CongruenceLattice con;

  // depends on alpha:
  private Partition alpha;
  private int[] roots;
  private int rootsSize;
  private HashSet diagonalHS;
  private ArrayList diagonal;
  private HashSet diagonal4HS;
  private ArrayList diagonal4;

  private HashSet typeSet;

//EWK
    //This maps each JI c to typ(c,c_*)
    private HashMap JItypes;

  public TypeFinder(CongruenceLattice congs) {
    con = congs;
    this.A = con.getAlgebra();
    algSize = A.getSize();
  }
//endEWK


  void setPartition(Partition alpha) {
    if (this.alpha == null || !this.alpha.equals(alpha)) {
      this.alpha = alpha;
      diagonalHS = new HashSet();
      diagonal = new ArrayList();
      diagonal4HS = new HashSet();
      diagonal4 = new ArrayList();
      this.alpha = alpha;
      roots = alpha.getRepresentatives();
      rootsSize = roots.length;
      for (int i = 0; i < rootsSize; i++) {
        int[] tmp = new int[] {roots[i], roots[i]};
        int[] tmp2 = new int[] {roots[i], roots[i], roots[i], roots[i]};
        diagonal.add(tmp);
        diagonalHS.add(new IntArray(tmp));
        diagonal4.add(tmp2);
        diagonal4HS.add(new IntArray(tmp2));
      }
    }
  }

//EWK
  public HashMap getJItypes() {
      return JItypes;
  }
//endEWK

  public void makeJIs() {
    if (showTime) {
      long time = System.currentTimeMillis();
      con.makePrincipals();
      con.makeJICongruences();
      System.out.println("time to make principals and join irreducibles was " 
                               + (System.currentTimeMillis() - time));
    }
    else {
      con.makePrincipals();
      con.makeJICongruences();
    }
  }

  /**
   * Find the TCT type set of the algebra A.
   */
  public HashSet findTypeSet() {
    if (typeSet != null) return typeSet;
    else typeSet = new HashSet();
//EWK
    if (con.getJoinIrreducibles() == null) makeJIs();
    JItypes = new HashMap(con.getJoinIrreducibles().size());
//endEWK

/*
    HashSet tested = new HashSet();
    for (int i = 0; i < algSize; i++) {
      for (int j = i+1; j < algSize; j++) {
        Partition cgab = con.CgPartition(i, j);
        if (!tested.contains(cgab)) {
          tested.add(cgab);
          Partition cgablc = con.lowerStarPartition(cgab);
          if (cgablc != null) {
            int typ = findType(i, j, cgablc);
            typeSet.add(new Integer(typ));
          }
        }
      }
    }
*/
//System.out.println("the number of jis is " + con.getJoinIrreducibles().size());
    Iterator it = con.getJoinIrreducibles().iterator();
    while (it.hasNext()) {
      Partition ji = (Partition)it.next();
      Partition jilc = con.lowerStarPartition(ji);
      int[] gen = con.getGeneratingPair(ji);

      Integer type = new Integer(findType(gen[0], gen[1], jilc));
//EWK
      JItypes.put(ji,type);
//endEWK
      typeSet.add(type);

    }
    return typeSet;
  }

  /**
   * Finds the type of Cg(a,b) over its lower cover.
   * Cg(a,b) must be join irreducible and alpha must be a congruence
   * which is not above Cg(a,b) but is above the lower cover.
   */
  public int findType(int a, int b, Partition alpha) {
    int[] subtr = findSubtrace(a, b, alpha);  // this sets alpha too.
    //return findType(subtr, hasInvolution(subtr));
    return findType(subtr, hasInvolution());
  }

  /**
   * If beta covers alpha, this finds the type of the quotient. As
   * long as beta is not below alpha, it finds the type of some gamma
   * below beta covering alpha meet beta.
   */
  public int findType(Partition alpha, Partition beta) {
//EWK
    if (con.getJoinIrreducibles() == null) makeJIs();
//endEWK
    beta = con.findJoinIrred(alpha, beta);
    int[] gen = con.getGeneratingPair(beta);
    int[] subtr = findSubtrace(gen[0], gen[1], alpha);  // this sets alpha too.
    //return findType(subtr, hasInvolution(subtr));
    return findType(subtr, hasInvolution());
  }


  /**
   * Find a subtrace [c, d] such that Cg(c, d) = Cg(a, b) which is 
   * assumed to be join irreducible.
   */
  public int[] findSubtrace(int a, int b, Partition alpha) {
    visited = new HashSet();
    setPartition(alpha);
    a = alpha.representative(a);
    b = alpha.representative(b);
    // we should throw an exception if a now equals b
    if (b < a) {
      int c = a;
      a = b;
      b = c;
    }
    hasInvolution = false;
    //return findSubtrace(new IntArray(new int[] {a, b}));
    int[] subtr = findSubtrace(new IntArray(new int[] {a, b}));

//System.out.println("subtrace " + (new IntArray(subtr)) + " has inv " + hasInvolution);

    return subtr;
  }

  private int[] findSubtrace(IntArray pairIA) {

    int[] pair = pairIA.getArray();
    visited.add(pairIA);
//First store the generators into universe.

    List universe = (List)diagonal.clone();

    HashSet genHashSet = (HashSet)diagonalHS.clone(); 
    genHashSet.add(pairIA);
    universe.add(pair);

//We do the generation in two steps.
//In fact lots of the code for the two steps is identical,
//but I want to avoid an "if" in the middle of a tight loop.
//In Step 1, we apply all operations to all vectors in universe.
//The resulting vectors are put into "newest".

    int i, j, k;
    ArrayList newest = new ArrayList();
    Iterator op_it = A.getOperationsIterator();
    // holding places for the result. *Note* these are fixed places 
    // that get constantly overwritten so we need to copy them if we
    // want to save them. !!!!
    IntArray resIA = new IntArray(2);
    int[] res = resIA.getArray();    // Note writting to res changes resIA
    // a separate holder for the unordered pair
    IntArray unorderedResIA = new IntArray(2);
    int[] unorderedRes = unorderedResIA.getArray(); // Note  (see above)
    while( op_it.hasNext() ) {//through operations
      Operation f = (Operation)op_it.next();
      int ar = f.getArity();
      Iterator arg_it[] = new Iterator[ar];
      // ArrayList arg[] = new ArrayList[ar];
      int[][] arg = new int[ar][];
      int argv[] = new int[ar];
      for(j=0; j<ar; j++) {
        arg_it[j] = universe.iterator();
        // arg[j] = (ArrayList)arg_it[j].next();
        arg[j] = (int[])arg_it[j].next();
      }
      //This assumes that universe is not empty.
      while(true) {//through arguments
        //perform operation f:
        for(k=0; k < 2; k++ ) {
          for( j=0; j < ar; j++) {
            // argv[j] = ((Integer)arg[j].get(k)).intValue();
            //argv[j] = arg[j].get(k);
            argv[j] = arg[j][k];
          }
          res[k] = alpha.representative(f.getValue(argv)); // normalizes
        }
        // ignore res if it is in alpha:
        if (res[0] != res[1]) {
          if (res[1] < res[0]) {
            unorderedRes[0] = res[1];
            unorderedRes[1] = res[0];
          }
          else {
            unorderedRes[0] = res[0];
            unorderedRes[1] = res[1];
          }
          if (!visited.contains(unorderedResIA)) {
            return findSubtrace((IntArray)unorderedResIA.clone());
          }
          else {
            if ( !genHashSet.contains( resIA ) ) {
              int[] res2 = new int[2];
              for (int v = 0; v < 2; v++) {
                res2[v] = res[v];
              }
              genHashSet.add(new IntArray(res2));
              newest.add( res2 );
            }
          }
        }
    
        //increment the argumentlist
        for(j=0; j< ar; j++ ) {
          if( !arg_it[j].hasNext() ) {
            arg_it[j] = universe.iterator();
            //arg[j] = (ArrayList)arg_it[j].next();
            //arg[j] = (IntArray)arg_it[j].next();
            arg[j] = (int[])arg_it[j].next();
            continue;//increment next coordinate
          } 
          // arg[j] = (ArrayList)arg_it[j].next();
          // arg[j] = (IntArray)arg_it[j].next();
          arg[j] = (int [])arg_it[j].next();
          break;//arg_it[j] has been increased
        }
        if ( j == ar ) {//all arguments done
          break;
        }
      }//arguments cycle
    }//operations cycle
    if( newest.isEmpty() ) {//we are done
      if (genHashSet.contains(new IntArray(new int[] {pair[1], pair[0]}))) {
        hasInvolution = true;
      }
      return pair;
    }

//Now comes Step 2. 
//We keep three ArrayLists:
//"old" is the beginning of the generated vectors,
//  where all values have been checked;
//"lastNew" is the list of vectors computed last time;
//universe is always the join of old and lastNew.
//All this is to avoid computing the value of an
//operation twice on the same arguments.

    ArrayList old;
    ArrayList lastNew;
    while(true) {//through "levels"
      old = new ArrayList(universe);
      //It is a waste of time to copy universe.
      //However, if we don't, then there is now way
      //I can tell the iterator that it should go only
      //up to the end of the old list. And I assume
      //that it is much slower to avoid iterators in the code.


//System.out.println(universe.size());
      Iterator it = newest.iterator();
      while( it.hasNext() ) {
        // universe.add((ArrayList)it.next());
        // universe.add((IntArray)it.next());
        universe.add((int[])it.next());
      }
      lastNew = newest;
      newest = new ArrayList();
      op_it = A.getOperationsIterator();
      while( op_it.hasNext() ) {//through operations
        Operation f = (Operation)op_it.next();
        int ar = f.getArity();
        Iterator arg_it[] = new Iterator[ar];
        // ArrayList arg[] = new ArrayList[ar];
        // IntArray arg[] = new IntArray[ar];
        int[][] arg = new int[ar][];
        int argv[] = new int[ar];
        for(i=0; i<ar; i++) {
          //Perform operation f for the arguments
          //given in Iterator[]. 
          //In the i-th step, 
          //  the first i arguments are in old;
          //  the i-th argument is in lastNew;
          //  the rest of the arguments is in universe
          for(j=0; j<i; j++) {
            arg_it[j] = old.iterator();
            // arg[j] = (ArrayList)arg_it[j].next();
            // arg[j] = (IntArray)arg_it[j].next();
            arg[j] = (int[])arg_it[j].next();
          }
          arg_it[i] = lastNew.iterator();
          //arg[i] = (ArrayList)arg_it[i].next();
          //arg[i] = (IntArray)arg_it[i].next();
          arg[i] = (int[])arg_it[i].next();
          for(j=i+1; j<ar; j++) {
            arg_it[j] = universe.iterator();
            // arg[j] = (ArrayList)arg_it[j].next();
            // arg[j] = (IntArray)arg_it[j].next();
            arg[j] = (int[])arg_it[j].next();
          }
          //This assumes that "old" and lastNew
          //and universe are not empty.
          // rsf: I ;think the next line is unnecessary so I'm deleting it.
          // res = new int[2];
          while(true) {//through arguments
            //perform operation f:
            for(k=0; k<2; k++ ) {
              for( j=0; j< ar; j++) {
                argv[j] = arg[j][k];
              }
              res[k] = alpha.representative(f.getValue(argv));
            }
            // ignore res if it is in alpha:
            if (res[0] != res[1]) {
              if (res[1] < res[0]) {
                unorderedRes[0] = res[1];
                unorderedRes[1] = res[0];
              }
              else {
                unorderedRes[0] = res[0];
                unorderedRes[1] = res[1];
              }
              if (!visited.contains(unorderedResIA)) {
                return findSubtrace((IntArray)unorderedResIA.clone());
              }
              else {
                if( !genHashSet.contains( resIA ) ) {
                  int [] res2 = new int[2];
                  for (int v = 0; v < 2; v++) {
                    res2[v] = res[v];
                  }
                  genHashSet.add(new IntArray(res2));
                  newest.add( res2 );
                }
              }
            }

            //increment the argumentlist
            for(j=0; j< ar; j++ ) {
              if( !arg_it[j].hasNext() ) {
                if( j < i ) {
                  arg_it[j] = old.iterator();
                } else if( j == i ) {
                  arg_it[j] = lastNew.iterator();
                } else {
                  arg_it[j] = universe.iterator();
                }
                // arg[j] = (ArrayList)arg_it[j].next();
                // arg[j] = (IntArray)arg_it[j].next();
                arg[j] = (int[])arg_it[j].next();
                continue;//increment next coordinate
              } 
              // arg[j] = (ArrayList)arg_it[j].next();
              // arg[j] = (IntArray)arg_it[j].next();
              arg[j] = (int[])arg_it[j].next();
              break;//arg_it[j] has been increased
            }
            if ( j == ar ) {//all arguments done
              break;
            }
          }//arguments cycle
        }
      }//operations cycle
      if( newest.isEmpty() ) {//we are done
        if (genHashSet.contains(new IntArray(new int[] {pair[1], pair[0]}))) {
          hasInvolution = true;
        }
        return pair;
      }
    }
  }

  /**
   * This determines if the last subtrace found by findSubtrace
   * has an involution. It will return false if findSubtrace
   * not yet been run.
   */
  private boolean hasInvolution() { 
    return hasInvolution; 
  }

  /**
   * Finds the type of the subtrace found with findSubtrace.
   * <code>hasInvolution</code> should have been set.
   * <p>
   * Let [c,d] be answer found by findSubtrace. We think of quadruples 
   * as maps from {c,d}^2 into A in the row order. That is
   * (c,c), (c,d), (d,c), (d,d). So the projections are [c,c,d,d] and
   * [c,d,c,d].  So a (c,d) 2-snag is [c,c,c,d] and a (d,c) 2-snag is
   * [c,d,d,d].
   * 
   * 
   * 
   */
  int findType(int[] subtrace, boolean hasInvolution) {

long time = System.currentTimeMillis();



    boolean has2snag = false;      // (c,d) is a 2-snag
    boolean has2snagRev = false;   // (d,c) is a 2-snag
    boolean has1snag = false;

//EWK
    boolean meet = false;
    boolean join=false;
    boolean oneSnag = false;
    int wx,wy,wu,wv;
//    int count=0;
//endEWK
   

    List universe = (List)diagonal4.clone();
    HashSet genHashSet = (HashSet)diagonal4HS.clone(); 
    int c = subtrace[0];
    int d = subtrace[1];

//    System.out.println(c + "," + d);



    int[] tmp = new int[] {c,c,d,d};
    universe.add(tmp);
    genHashSet.add(new IntArray(tmp));
    tmp = new int[] {c,d,c,d};
    universe.add(tmp);
    genHashSet.add(new IntArray(tmp));

//We do the generation in two steps.
//In fact lots of the code for the two steps is identical,
//but I want to avoid an "if" in the middle of a tight loop.
//In Step 1, we apply all operations to all vectors in universe.
//The resulting vectors are put into "newest".

    int i, j, k;
    ArrayList newest = new ArrayList();
    Iterator op_it = A.getOperationsIterator();
    while( op_it.hasNext() ) {//through operations
      Operation f = (Operation)op_it.next();
      int ar = f.getArity();
      Iterator arg_it[] = new Iterator[ar];
      int[][] arg = new int[ar][];
      int argv[] = new int[ar];
      for(j=0; j<ar; j++) {
        arg_it[j] = universe.iterator();
        arg[j] = (int[])arg_it[j].next();
      }
      //This assumes that universe is not empty.
      int[] res = new int[4];
      IntArray resIA = new IntArray(4);
      while(true) {//through arguments
        //perform operation f:
        for(k=0; k<4; k++ ) {
          for( j=0; j< ar; j++) {
            argv[j] = arg[j][k];
          }
          res[k] = f.getValue(argv);
        }
        // normalize:
        res[0] = alpha.representative(res[0]);
        res[1] = alpha.representative(res[1]);
        res[2] = alpha.representative(res[2]);
        res[3] = alpha.representative(res[3]);
        resIA.setIntArray(res);
        if( !genHashSet.contains( resIA ) ) {
          int [] res2 = new int[4];
          for (int v = 0; v < 4; v++) {
            res2[v] = res[v];
          }
          genHashSet.add(new IntArray(res2));
          newest.add( res2 );
//EWK
//	  count++;
	  wx=res2[0]; wy=res2[1]; wu=res2[2]; wv= res2[3];
//	  System.out.println(wx + "," + wy + "," + wu + "," + wv);
	  if( ((wx!=wy) && (wu==wv)) ||
	      ((wx!=wu) && (wy==wv)) ) { /* join found */
	      if( hasInvolution ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 3;
	      }
	      if( meet ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 4;
	      }
	      join = true;
	      oneSnag = true;
	  } else if( ((wx==wy) && (wu!=wv)) ||
		     ((wx==wu) && (wy!=wv)) ) { /* meet found */
	      if( hasInvolution ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 3;
	      }
	      if( join ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 4;
	      }
	      meet = true;
	      oneSnag = true;
	  }
	  //So not a meet, not a join. Is it an other kind of one-snag?
	  if( !oneSnag ) {
	      if(  ( (wx==wv) && ((wx!=wy) || (wu!=wv)) ) ||
		   ( (wy==wu) && ((wx!=wy) || (wu!=wv)) )    ) {
		  oneSnag=true;
	      }
	  }
//endEWK
        }
    
        //increment the argumentlist
        for(j=0; j< ar; j++ ) {
          if( !arg_it[j].hasNext() ) {
            arg_it[j] = universe.iterator();
            arg[j] = (int[])arg_it[j].next();
            continue;//increment next coordinate
          } 
          arg[j] = (int [])arg_it[j].next();
          break;//arg_it[j] has been increased
        }
        if ( j == ar ) {//all arguments done
          break;
        }
      }//arguments cycle
    }//operations cycle
    if( newest.isEmpty() ) {//we are done
//	  System.out.println(count);
	  type_time += System.currentTimeMillis() - time;
	  if(join==true || meet==true) {
	      return 5;
	  } else {
	      return (oneSnag==true) ? 2 : 1;
	  }
    }

//Now comes Step 2. 
//We keep three ArrayLists:
//"old" is the beginning of the generated vectors,
//  where all values have been checked;
//"lastNew" is the list of vectors computed last time;
//universe is always the join of old and lastNew.
//All this is to avoid computing the value of an
//operation twice on the same arguments.

    ArrayList old;
    ArrayList lastNew;
    while(true) {//through "levels"
      old = new ArrayList(universe);
      //It is a waste of time to copy universe.
      //However, if we don't, then there is now way
      //I can tell the iterator that it should go only
      //up to the end of the old list. And I assume
      //that it is much slower to avoid iterators in the code.


//System.out.println(universe.size());
      Iterator it = newest.iterator();
      while( it.hasNext() ) {
        universe.add((int[])it.next());
      }
      lastNew = newest;
      newest = new ArrayList();
      op_it = A.getOperationsIterator();
      while( op_it.hasNext() ) {//through operations
        Operation f = (Operation)op_it.next();
        int ar = f.getArity();
        Iterator arg_it[] = new Iterator[ar];
        int[][] arg = new int[ar][];
        int argv[] = new int[ar];
        for(i=0; i<ar; i++) {
          //Perform operation f for the arguments
          //given in Iterator[]. 
          //In the i-th step, 
          //  the first i arguments are in old;
          //  the i-th argument is in lastNew;
          //  the rest of the arguments is in universe
          for(j=0; j<i; j++) {
            arg_it[j] = old.iterator();
            arg[j] = (int[])arg_it[j].next();
          }
          arg_it[i] = lastNew.iterator();
          arg[i] = (int[])arg_it[i].next();
          for(j=i+1; j<ar; j++) {
            arg_it[j] = universe.iterator();
            arg[j] = (int[])arg_it[j].next();
          }
          //This assumes that "old" and lastNew
          //and universe are not empty.
          int[] res = new int[4];
          while(true) {//through arguments
            //perform operation f:
            for(k=0; k<4; k++ ) {
              for( j=0; j< ar; j++) {
                argv[j] = arg[j][k];
              }
              res[k] = f.getValue(argv);
            }
            // normalize:
            res[0] = alpha.representative(res[0]);
            res[1] = alpha.representative(res[1]);
            res[2] = alpha.representative(res[2]);
            res[3] = alpha.representative(res[3]);
            IntArray resIA = new IntArray(res);
            if( !genHashSet.contains( resIA ) ) {
              int[] res2 = new int[4];
              for (int v = 0; v < 4; v++) {
                res2[v] = res[v];
              }
              genHashSet.add(new IntArray(res2));
              newest.add( res2 );
//EWK

//	  count++;
	  wx=res2[0]; wy=res2[1]; wu=res2[2]; wv= res2[3];
//	  System.out.println(wx + "," + wy + "," + wu + "," + wv);
	  if( ((wx!=wy) && (wu==wv)) ||
	      ((wx!=wu) && (wy==wv)) ) { /* join found */
	      if( hasInvolution ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 3;
	      }
	      if( meet ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 4;
	      }
	      join = true;
	      oneSnag = true;
	  } else if( ((wx==wy) && (wu!=wv)) ||
		     ((wx==wu) && (wy!=wv)) ) { /* meet found */
	      if( hasInvolution ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 3;
	      }
	      if( join ) {
//		  System.out.println(count);
		  type_time += System.currentTimeMillis() - time;
		  return 4;
	      }
	      meet = true;
	      oneSnag = true;
	  }
	  //So not a meet, not a join. Is it an other kind of one-snag?
	  if( !oneSnag ) {
	      if(  ( (wx==wv) && ((wx!=wy) || (wu!=wv)) ) ||
		   ( (wy==wu) && ((wx!=wy) || (wu!=wv)) )    ) {
		  oneSnag=true;
	      }
	  }

//endEWK
            }

            //increment the argumentlist
            for(j=0; j< ar; j++ ) {
              if( !arg_it[j].hasNext() ) {
                if( j < i ) {
                  arg_it[j] = old.iterator();
                } else if( j == i ) {
                  arg_it[j] = lastNew.iterator();
                } else {
                  arg_it[j] = universe.iterator();
                }
                arg[j] = (int[])arg_it[j].next();
                continue;//increment next coordinate
              } 
              arg[j] = (int[])arg_it[j].next();
              break;//arg_it[j] has been increased
            }
            if ( j == ar ) {//all arguments done
              break;
            }
          }//arguments cycle
        }
      }//operations cycle
      if( newest.isEmpty() ) {//we are done
//EWK

//	  System.out.println(count);
	  type_time += System.currentTimeMillis() - time;
	  if(join==true || meet==true) {
	      return 5;
	  } else {
	      return (oneSnag==true) ? 2 : 1;
	  }

//endEWK
      }
    }
  }



  public static void main(String[] args) throws Exception {
    if (args.length < 1) {
      System.out.println("give an alg file and 2 ints in the alg");
      return;
    }
    Algebra alg = new Algebra("test");
    ewkInputStream fp = new ewkInputStream(new FileInputStream(args[0]));
    alg.Read(fp);
    if (fp != null) fp.close();

/*
    CongruenceLattice con = new CongruenceLattice(alg);
//    con.computeAll();
    long time = System.currentTimeMillis();
    con.makePrincipals();
    long time2 = System.currentTimeMillis();
    System.out.println("made principals, time: " + (time2 - time));
    time = time2;
    con.makeJICongruences();
    time2 = System.currentTimeMillis();
    System.out.println("made ji's, time: " + (time2 - time));
    //System.out.println("con has size " + con.getSize());
*/
    CongruenceLattice c = new CongruenceLattice(alg);
    TypeFinder subtr = new TypeFinder(c);
    System.out.println("\nFinding the type set of A");
    long time = System.currentTimeMillis();
    subtr.findTypeSet();
    time = System.currentTimeMillis() - time;
    System.out.print("typeSet :");
    for (int i=1; i < 6; i++) {
      if (subtr.typeSet.contains(new Integer(i))) System.out.print(" " + i);
    }
    System.out.println("");
    System.out.println("the compute time is " + time);
    System.out.println("the type computing time is " + type_time);
    
  }

    private static long type_time = 0;

}

