package uacalc.alg.opedit;

import uacalc.alg.Algebra;
import uacalc.alg.Operation;
import uacalc.alg.OpArguments;
import uacalc.util.WholeNumberField;
import uacalc.config.Configure;

import uacalc.ex.TooManyArgumentsException;

import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

public class OperationRendererTable
    extends OperationRenderer {

    public OperationRendererTable(Operation f,
				  ValueEditor editor,
				  Container pane) {
	super(f, editor, pane);

	tableModel = new OpValueTableModel();
	OpValueTable = 
	    new ValueEditorTable(m_f, tableModel);

	OpValueTable.setTableCursorAt(0, 0);

	m_pane.setLayout(new BoxLayout(m_pane, BoxLayout.X_AXIS));
	m_pane.add(OpValueTable);
    }

    public void editValueAt(OpArguments args, int value) {
	OpValueTable.editCellAt(args.getValue(0),
			       args.getValue(1),
				value);
    }

    public void editValueAt(OpArguments args) {
	OpValueTable.editCellAt(args.getValue(0),
			       args.getValue(1), -2);
    }

    public void Update() {
	tableModel.fireTableDataChanged();
	OpValueTable.setFocusToEditedCell();
    }

    public boolean isEditorOpen() {
	return OpValueTable.isEditorOpen();
    }

    private ValueEditorTable OpValueTable;
    private OpValueTableModel tableModel;

//inner classes

    class OpValueTableModel extends AbstractTableModel {

        public int getColumnCount() {
            return m_size;
        }

	public int getRowCount() {
            return m_size;
        }

	public String getColumnName(int col) {
            return Integer.toString(col);
	}

	public Object getValueAt(int row, int col) {
            return new Integer(m_f.getValue(row, col));
        }

	public void setValueAt(Object value, int row, int col) {

	    OpArguments args;
	    try {
		args = new OpArguments(m_arity);
	    } catch(TooManyArgumentsException e) {
		System.out.println(e.toString());
		return;
	    }
	    args.setValue(0, row);
	    args.setValue(1, col);
	    int intval = ((Integer)value).intValue();
	    if( !m_f.tryValue(row, col, intval ) ) {
		OnInvalidValue( args, intval );
	    }
	    fireTableCellUpdated(row, col);
	    OperationChanged();
	}


	public Class getColumnClass(int col) {
            return Integer.class;
        }

	public boolean isCellEditable(int row, int col) {
	    return true;
        }
    }
   
}


class ValueEditorTable extends JScrollPane {

    public ValueEditorTable(Operation f, 
			    AbstractTableModel model) {
//ONLY FOR BINARY f!

	m_f = f;
	m_model = model;
	m_size = f.getSize();
	int i;

        table = new JTable(m_model);

//set up value fields
	NumberField =  
	    new WholeNumberField(0, Algebra.MAX_ALG_SIZE_DIGITS);
        DefaultCellEditor integerEditor =
            new DefaultCellEditor(NumberField) {
		    //Override DefaultCellEditor's getCellEditorValue method
		    //to return an Integer, not a String:
		    public Object getCellEditorValue() {
			return new Integer(NumberField.getValue());
		    }
		};
	
        table.setDefaultEditor(Integer.class, integerEditor);

	table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	table.setCellSelectionEnabled(true);
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	table.getTableHeader().setReorderingAllowed(false);
	table.getTableHeader().setResizingAllowed(false);

//Row Headers are in a separate table
	rowHeaderTableModel rowHeaderModel = new rowHeaderTableModel();
        rowHeaderTable = new JTable(rowHeaderModel);

	rowHeaderTable.setTableHeader(null);
	rowHeaderTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	rowHeaderTable.setColumnSelectionAllowed(false);
	rowHeaderTable.setRowSelectionAllowed(false);
//	rowHeaderTable.getTableHeader().setReorderingAllowed(false);
//	rowHeaderTable.getTableHeader().setResizingAllowed(false);

//set up cell dimensions:
	int width = (int)NumberField.getPreferredSize().getWidth();
	int height = (int)NumberField.getPreferredSize().getHeight();
	Dimension cellDim = new Dimension(width, height);

//cell width
	TableColumn column = null;
        for (i = 0; i < m_size; i++) {
            column = table.getColumnModel().getColumn(i);
	    column.setPreferredWidth(width);
	    column.setMaxWidth(width);
	    column.setMinWidth(width);
        }
	column = rowHeaderTable.getColumnModel().getColumn(0);
	column.setPreferredWidth(width);
	column.setMaxWidth(width);
	column.setMinWidth(width);

//cell height
	table.setRowHeight(height);
	rowHeaderTable.setRowHeight(height);

//set renderer for row and column headers
	rowHeaderTable.setDefaultRenderer(
	    rowHeaderModel.getColumnClass(0),
	    new HeaderRenderer(cellDim));

        for (i = 0; i < m_size; i++) {
            column = table.getColumnModel().getColumn(i);
	    column.setHeaderRenderer(
		new HeaderRenderer(cellDim));
        }

//put row header table into the scroll
	JViewport rowHeaderViewport = new JViewport();
	rowHeaderViewport.setView(rowHeaderTable);
	rowHeaderViewport.setPreferredSize(
	    rowHeaderTable.getPreferredSize());

//set up scrolling
	setViewportView(table);
	setRowHeaderView(rowHeaderViewport);
	setCorner(JScrollPane.UPPER_LEFT_CORNER,
			 new JLabel(m_f.getNameString(),
				    SwingConstants.CENTER));
	setMaxScrollSize();

//set preferred original size
	Dimension preferredTableSize = new Dimension(
	    Configure.DEFAULT_NUMBER_OF_VISIBLE_ELEMENTS*width, 
	    Configure.DEFAULT_NUMBER_OF_VISIBLE_ELEMENTS*height);
	table.setPreferredScrollableViewportSize(preferredTableSize);

//put it to the upper left corner:
	setAlignmentY(Component.TOP_ALIGNMENT);
    }

    public void setTableCursorAt(int row, int col) {

	table.requestFocus();
	table.setRowSelectionInterval(row, row);
	table.setColumnSelectionInterval(col, col);
    }

    public boolean isEditorOpen() {
	return table.isEditing();
    }

    public void setFocusToEditedCell() {

	WholeNumberField f = 
	    (WholeNumberField)(table.getEditorComponent());
	if ( f != null) f.requestFocus();
	else table.requestFocus();
    }


    public void editCellAt(int row, int col, int value) {

	table.scrollRectToVisible(
	    table.getCellRect(row, col, true));
	table.editCellAt(row, col);
	WholeNumberField f = 
	    (WholeNumberField)(table.getEditorComponent());
	if(value != -2 ) f.setValue(value);
	f.requestFocus();
	f.selectAll();
	table.setRowSelectionInterval(row, row);
	table.setColumnSelectionInterval(col, col);
    }

    private void setMaxScrollSize() {
	/** Includes: column and row headers,
	 *  scrollbars, borders.
	 *  needed for the BoxLayout manager
	 */

	int w;
	int h;

	w = (int)table.getMaximumSize().getWidth();
	h = (int)table.getMaximumSize().getHeight();
	w += (int)rowHeaderTable.getMaximumSize().getWidth();
	h += (int)table.getTableHeader().getMaximumSize().getHeight();
        w += (int)getVerticalScrollBar().getMinimumSize().getWidth();
        h += (int)getHorizontalScrollBar().getMinimumSize().getHeight();
	Insets ins = getBorder().getBorderInsets(this);
	w += ins.left + ins.right;
	h += ins.top + ins.bottom;
        Dimension MaxScrollDim = new Dimension();
	MaxScrollDim.setSize(w, h);
	setMaximumSize(MaxScrollDim);
    }


    private JTable table;
    private JTable rowHeaderTable;
    private Operation m_f;
    private int m_size;
    private WholeNumberField NumberField;
    private AbstractTableModel m_model;

//Inner classes

    class HeaderRenderer extends JLabel
	implements TableCellRenderer {
 
	HeaderRenderer(Dimension cellDim) {
	    JTable t = new JTable();
	    //Dummy table, to get default header attributes.

	    JTableHeader header = t.getTableHeader();
	    setOpaque(true);
	    setBorder(UIManager.getBorder("TableHeader.cellBorder"));
	    setHorizontalAlignment(CENTER);
	    setForeground(header.getForeground());
	    setBackground(header.getBackground());
	    setFont(header.getFont());
	    setPreferredSize(cellDim);
	    setMaximumSize(cellDim);
	    setMinimumSize(cellDim);
	}
 
	public Component 
	getTableCellRendererComponent(JTable table, Object value, 
				      boolean isSelected, boolean hasFocus,
				      int row, int column) {
	    setText((value == null) ? "" : value.toString());
	    return this;
	}
    }

    class rowHeaderTableModel extends AbstractTableModel {

        public int getColumnCount() {
            return 1;
        }

	public int getRowCount() {
            return m_size;
        }

	public Object getValueAt(int row, int col) {
            return Integer.toString(row);
        }

	public boolean isCellEditable(int row, int col) {
	    return false;
        }
    }

}

